# Copyright 2013, 2014 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import yaml

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils
from oslo_utils import uuidutils
from toscaparser.tosca_template import ToscaTemplate

from tacker._i18n import _
from tacker.api.v1 import attributes
from tacker.common import driver_manager
from tacker.common import exceptions
from tacker.common import utils
from tacker import context as t_context
from tacker.db.vnfm import vnfm_db
from tacker.extensions import vnfm
from tacker.plugins.common import constants
from tacker.plugins import fenix
from tacker.tosca import utils as toscautils
from tacker.vnfm.mgmt_drivers import constants as mgmt_constants
from tacker.vnfm import monitor
from tacker.vnfm import vim_client


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def config_opts():
    return [('tacker', VNFMMgmtMixin.OPTS),
            ('tacker', VNFMPlugin.OPTS_INFRA_DRIVER),
            ('tacker', VNFMPlugin.OPTS_POLICY_ACTION)]


class VNFMMgmtMixin(object):
    OPTS = [
        cfg.ListOpt(
            'mgmt_driver', default=['noop', 'openwrt'],
            help=_('MGMT driver to communicate with '
                   'Hosting VNF/logical service '
                   'instance tacker plugin will use')),
        cfg.IntOpt('boot_wait', default=30,
            help=_('Time interval to wait for VM to boot'))
    ]
    cfg.CONF.register_opts(OPTS, 'tacker')

    def __init__(self):
        super(VNFMMgmtMixin, self).__init__()
        self._mgmt_manager = driver_manager.DriverManager(
            'tacker.tacker.mgmt.drivers', cfg.CONF.tacker.mgmt_driver)

    def _invoke(self, vnf_dict, **kwargs):
        method = inspect.stack()[1][3]
        return self._mgmt_manager.invoke(
            self._mgmt_driver_name(vnf_dict), method, **kwargs)

    def mgmt_create_pre(self, context, vnf_dict):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict)

    def mgmt_create_post(self, context, vnf_dict):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict)

    def mgmt_update_pre(self, context, vnf_dict):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict)

    def mgmt_update_post(self, context, vnf_dict):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict)

    def mgmt_delete_pre(self, context, vnf_dict):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict)

    def mgmt_delete_post(self, context, vnf_dict):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict)

    def mgmt_get_config(self, context, vnf_dict):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict)

    def mgmt_ip_address(self, context, vnf_dict):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict)

    def mgmt_call(self, context, vnf_dict, kwargs):
        return self._invoke(
            vnf_dict, plugin=self, context=context, vnf=vnf_dict,
            kwargs=kwargs)


class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
    """VNFMPlugin which supports VNFM framework.

    Plugin which supports Tacker framework
    """

    OPTS_INFRA_DRIVER = [
        cfg.ListOpt(
            'infra_driver', default=['noop', 'openstack', 'kubernetes'],
            help=_('Hosting vnf drivers tacker plugin will use')),
    ]
    cfg.CONF.register_opts(OPTS_INFRA_DRIVER, 'tacker')

    OPTS_POLICY_ACTION = [
        cfg.ListOpt(
            'policy_action', default=['autoscaling', 'respawn',
                                      'vdu_autoheal', 'log', 'log_and_kill'],
            help=_('Hosting vnf drivers tacker plugin will use')),
    ]
    cfg.CONF.register_opts(OPTS_POLICY_ACTION, 'tacker')

    supported_extension_aliases = ['vnfm']

    def __init__(self):
        super(VNFMPlugin, self).__init__()
        self._pool = eventlet.GreenPool()
        self.boot_wait = cfg.CONF.tacker.boot_wait
        self.vim_client = vim_client.VimClient()
        self._vnf_manager = driver_manager.DriverManager(
            'tacker.tacker.vnfm.drivers',
            cfg.CONF.tacker.infra_driver)
        self._vnf_action = driver_manager.DriverManager(
            'tacker.tacker.policy.actions',
            cfg.CONF.tacker.policy_action)
        self._vnf_monitor = monitor.VNFMonitor(self.boot_wait)
        self._vnf_alarm_monitor = monitor.VNFAlarmMonitor()
        self._vnf_reservation_monitor = monitor.VNFReservationAlarmMonitor()
        self._vnf_maintenance_monitor = monitor.VNFMaintenanceAlarmMonitor()
        self._vnf_app_monitor = monitor.VNFAppMonitor()
        self._vnf_maintenance_plugin = fenix.FenixPlugin()
        self._init_monitoring()

    def _init_monitoring(self):
        context = t_context.get_admin_context()
        vnfs = self.get_vnfs(context)
        for vnf in vnfs:
            # Add tenant_id in context object as it is required
            # to get VIM in monitoring.
            context.tenant_id = vnf['tenant_id']
            self.add_vnf_to_monitor(context, vnf)

    def spawn_n(self, function, *args, **kwargs):
        self._pool.spawn_n(function, *args, **kwargs)

    def create_vnfd(self, context, vnfd):
        vnfd_data = vnfd['vnfd']
        template = vnfd_data['attributes'].get('vnfd')
        if isinstance(template, dict):
            # TODO(sripriya) remove this yaml dump once db supports storing
            # json format of yaml files in a separate column instead of
            # key value string pairs in vnf attributes table
            vnfd_data['attributes']['vnfd'] = yaml.safe_dump(
                template)
        else:
            raise vnfm.InvalidAPIAttributeType(atype=type(template))
        if "tosca_definitions_version" not in template:
            raise exceptions.Invalid('Not a valid template: '
                                     'tosca_definitions_version is missing.')

        LOG.debug('vnfd %s', vnfd_data)

        service_types = vnfd_data.get('service_types')
        if not attributes.is_attr_set(service_types):
            LOG.debug('service type must be specified')
            raise vnfm.ServiceTypesNotSpecified()
        for service_type in service_types:
            # TODO(yamahata):
            # framework doesn't know what services are valid for now.
            # so doesn't check it here yet.
            pass
        if 'template_source' in vnfd_data:
            template_source = vnfd_data.get('template_source')
        else:
            template_source = 'onboarded'
        vnfd['vnfd']['template_source'] = template_source

        self._parse_template_input(vnfd)
        return super(VNFMPlugin, self).create_vnfd(
            context, vnfd)

    def _parse_template_input(self, vnfd):
        vnfd_dict = vnfd['vnfd']
        vnfd_yaml = vnfd_dict['attributes'].get('vnfd')
        if vnfd_yaml is None:
            return

        inner_vnfd_dict = yaml.safe_load(vnfd_yaml)
        LOG.debug('vnfd_dict: %s', inner_vnfd_dict)

        # Prepend the tacker_defs.yaml import file with the full
        # path to the file
        toscautils.updateimports(inner_vnfd_dict)

        try:
            tosca = ToscaTemplate(a_file=False,
                                  yaml_dict_tpl=inner_vnfd_dict)
        except Exception as e:
            LOG.exception("tosca-parser error: %s", str(e))
            raise vnfm.ToscaParserFailed(error_msg_details=str(e))

        if ('description' not in vnfd_dict or
                vnfd_dict['description'] == ''):
            vnfd_dict['description'] = inner_vnfd_dict.get(
                'description', '')
        if (('name' not in vnfd_dict or
                not len(vnfd_dict['name'])) and
                'metadata' in inner_vnfd_dict):
            vnfd_dict['name'] = inner_vnfd_dict['metadata'].get(
                'template_name', '')

        vnfd_dict['mgmt_driver'] = toscautils.get_mgmt_driver(
            tosca)

        if vnfd_dict['mgmt_driver'] not in cfg.CONF.tacker.mgmt_driver:
            LOG.error("Invalid mgmt_driver in TOSCA template")
            raise vnfm.InvalidMgmtDriver(
                mgmt_driver_name=vnfd_dict['mgmt_driver'])

        LOG.debug('vnfd %s', vnfd)

    def add_vnf_to_monitor(self, context, vnf_dict):
        dev_attrs = vnf_dict['attributes']
        mgmt_ip_address = vnf_dict['mgmt_ip_address']
        if 'monitoring_policy' in dev_attrs and mgmt_ip_address:
            def action_cb(action, **kwargs):
                LOG.debug('policy action: %s', action)
                self._vnf_action.invoke(
                    action, 'execute_action', plugin=self, context=context,
                    vnf_dict=hosting_vnf['vnf'], args=kwargs)

            hosting_vnf = self._vnf_monitor.to_hosting_vnf(
                vnf_dict, action_cb)
            LOG.debug('hosting_vnf: %s', hosting_vnf)
            self._vnf_monitor.add_hosting_vnf(hosting_vnf)

    def add_alarm_url_to_vnf(self, context, vnf_dict):
        vnfd_yaml = vnf_dict['vnfd']['attributes'].get('vnfd', '')
        vnfd_dict = yaml.safe_load(vnfd_yaml)
        if not (vnfd_dict and vnfd_dict.get('tosca_definitions_version')):
            return
        try:
            toscautils.updateimports(vnfd_dict)
            tosca_vnfd = ToscaTemplate(a_file=False,
                                  yaml_dict_tpl=vnfd_dict)
        except Exception as e:
            LOG.exception("tosca-parser error: %s", str(e))
            raise vnfm.ToscaParserFailed(error_msg_details=str(e))
        polices = vnfd_dict['topology_template'].get('policies', [])
        for policy_dict in polices:
            name, policy = list(policy_dict.items())[0]
            if policy['type'] in constants.POLICY_ALARMING:
                alarm_url =\
                    self._vnf_alarm_monitor.update_vnf_with_alarm(
                        self, context, vnf_dict, policy)
                vnf_dict['attributes']['alarming_policy'] = vnf_dict['id']
                vnf_dict['attributes'].update(alarm_url)
            elif policy['type'] in constants.POLICY_RESERVATION:
                alarm_url = \
                    self._vnf_reservation_monitor.update_vnf_with_reservation(
                        self, context, vnf_dict, policy)
                vnf_dict['attributes']['reservation_policy'] = vnf_dict['id']
                vnf_dict['attributes'].update(alarm_url)
        maintenance_vdus = toscautils.find_maintenance_vdus(tosca_vnfd)
        maintenance = \
            self._vnf_maintenance_monitor.update_vnf_with_maintenance(
                vnf_dict, maintenance_vdus)
        vnf_dict['attributes'].update({
            'maintenance': jsonutils.dumps(maintenance['vdus'])})
        vnf_dict['attributes']['maintenance_url'] = maintenance['url']

    def add_vnf_to_appmonitor(self, context, vnf_dict):
        appmonitor = self._vnf_app_monitor.create_app_dict(context, vnf_dict)
        self._vnf_app_monitor.add_to_appmonitor(appmonitor, vnf_dict)

    def config_vnf(self, context, vnf_dict):
        config = vnf_dict['attributes'].get('config')
        if not config:
            self._vnf_maintenance_plugin.create_vnf_constraints(self, context,
                                                                vnf_dict)
            return
        if isinstance(config, str):
            # TODO(dkushwaha) remove this load once db supports storing
            # json format of yaml files in a separate column instead of
            #  key value string pairs in vnf attributes table.
            config = yaml.safe_load(config)

        eventlet.sleep(self.boot_wait)      # wait for vm to be ready
        vnf_id = vnf_dict['id']
        update = {
            'vnf': {
                'id': vnf_id,
                'attributes': {'config': config},
            }
        }
        self.update_vnf(context, vnf_id, update)

    def _get_infra_driver(self, context, vnf_info):
        vim_res = self.get_vim(context, vnf_info)
        return vim_res['vim_type'], vim_res['vim_auth']

    def _create_vnf_wait(self, context, vnf_dict, auth_attr, driver_name):
        vnf_id = vnf_dict['id']
        instance_id = self._instance_id(vnf_dict)
        create_failed = False

        try:
            self._vnf_manager.invoke(
                driver_name, 'create_wait', plugin=self, context=context,
                vnf_dict=vnf_dict, vnf_id=instance_id,
                auth_attr=auth_attr)
        except vnfm.VNFCreateWaitFailed as e:
            LOG.error("VNF Create failed for vnf_id %s", vnf_id)
            create_failed = True
            vnf_dict['status'] = constants.ERROR
            self.set_vnf_error_status_reason(context, vnf_id, str(e))

        if instance_id is None or create_failed:
            mgmt_ip_address = None
        else:
            # mgmt_ip_address = self.mgmt_ip_address(context, vnf_dict)
            # FIXME(yamahata):
            mgmt_ip_address = vnf_dict['mgmt_ip_address']

        self._create_vnf_post(
            context, vnf_id, instance_id, mgmt_ip_address, vnf_dict)
        self.mgmt_create_post(context, vnf_dict)

        if instance_id is None or create_failed:
            return

        vnf_dict['mgmt_ip_address'] = mgmt_ip_address

        kwargs = {
            mgmt_constants.KEY_ACTION: mgmt_constants.ACTION_CREATE_VNF,
            mgmt_constants.KEY_KWARGS: {'vnf': vnf_dict},
        }
        new_status = constants.ACTIVE
        try:
            self.mgmt_call(context, vnf_dict, kwargs)
        except exceptions.MgmtDriverException:
            LOG.error('VNF configuration failed')
            new_status = constants.ERROR
            self.set_vnf_error_status_reason(context, vnf_id,
                                             'Unable to configure VDU')
        vnf_dict['status'] = new_status
        self._create_vnf_status(context, vnf_id, new_status)

    def get_vim(self, context, vnf):
        region_name = vnf.setdefault('placement_attr', {}).get(
            'region_name', None)
        vim_res = self.vim_client.get_vim(context, vnf['vim_id'],
                                          region_name)
        vnf['placement_attr']['vim_name'] = vim_res['vim_name']
        vnf['vim_id'] = vim_res['vim_id']
        return vim_res

    def _create_vnf(self, context, vnf, vim_auth, driver_name):
        vnf_dict = self._create_vnf_pre(
            context, vnf) if not vnf.get('id') else vnf
        vnf_id = vnf_dict['id']
        LOG.debug('vnf_dict %s', vnf_dict)
        if driver_name == 'openstack':
            self.mgmt_create_pre(context, vnf_dict)
            self.add_alarm_url_to_vnf(context, vnf_dict)
        vnf_dict['attributes']['maintenance_group'] = uuidutils.generate_uuid()

        try:
            instance_id = self._vnf_manager.invoke(
                driver_name, 'create', plugin=self,
                context=context, vnf=vnf_dict, auth_attr=vim_auth)
        except Exception:
            LOG.debug('Fail to create vnf %s in infra_driver, '
                      'so delete this vnf',
                      vnf_dict['id'])
            with excutils.save_and_reraise_exception():
                self._delete_vnf(context, vnf_id, force_delete=True)

        vnf_dict['instance_id'] = instance_id
        return vnf_dict

    def create_vnf(self, context, vnf):
        vnf_info = vnf['vnf']
        name = vnf_info['name']

        # if vnfd_template specified, create vnfd from template
        # create template dictionary structure same as needed in create_vnfd()
        if vnf_info.get('vnfd_template'):
            vnfd_name = utils.generate_resource_name(name, 'inline')
            vnfd = {'vnfd': {'attributes': {'vnfd': vnf_info['vnfd_template']},
                             'name': vnfd_name,
                             'template_source': 'inline',
                             'service_types': [{'service_type': 'vnfd'}]}}
            vnf_info['vnfd_id'] = self.create_vnfd(context, vnfd).get('id')

        infra_driver, vim_auth = self._get_infra_driver(context, vnf_info)
        if infra_driver not in self._vnf_manager:
            LOG.debug('unknown vim driver '
                      '%(infra_driver)s in %(drivers)s',
                      {'infra_driver': infra_driver,
                       'drivers': cfg.CONF.tacker.infra_driver})
            raise vnfm.InvalidInfraDriver(vim_name=infra_driver)

        vnf_attributes = vnf_info['attributes']
        if vnf_attributes.get('param_values'):
            param = vnf_attributes['param_values']
            if isinstance(param, dict):
                # TODO(sripriya) remove this yaml dump once db supports storing
                # json format of yaml files in a separate column instead of
                #  key value string pairs in vnf attributes table
                vnf_attributes['param_values'] = yaml.safe_dump(param)
            else:
                raise vnfm.InvalidAPIAttributeType(atype=type(param))
        if vnf_attributes.get('config'):
            config = vnf_attributes['config']
            if isinstance(config, dict):
                # TODO(sripriya) remove this yaml dump once db supports storing
                # json format of yaml files in a separate column instead of
                #  key value string pairs in vnf attributes table
                vnf_attributes['config'] = yaml.safe_dump(config)
            else:
                raise vnfm.InvalidAPIAttributeType(atype=type(config))

        vnf_dict = self._create_vnf(context, vnf_info, vim_auth, infra_driver)

        def create_vnf_wait():
            self._create_vnf_wait(context, vnf_dict, vim_auth, infra_driver)

            if 'app_monitoring_policy' in vnf_dict['attributes']:
                self.add_vnf_to_appmonitor(context, vnf_dict)
            if vnf_dict['status'] is not constants.ERROR:
                self.add_vnf_to_monitor(context, vnf_dict)

            self.config_vnf(context, vnf_dict)
        self.spawn_n(create_vnf_wait)
        return vnf_dict

    # not for wsgi, but for service to create hosting vnf
    # the vnf is NOT added to monitor.
    def create_vnf_sync(self, context, vnf):
        infra_driver, vim_auth = self._get_infra_driver(context, vnf)
        vnf_dict = self._create_vnf(context, vnf, vim_auth, infra_driver)
        self._create_vnf_wait(context, vnf_dict, vim_auth, infra_driver)
        return vnf_dict

    def _heal_vnf_wait(self, context, vnf_dict, vim_auth, driver_name):
        kwargs = {
            mgmt_constants.KEY_ACTION: mgmt_constants.ACTION_HEAL_VNF,
            mgmt_constants.KEY_KWARGS: {'vnf': vnf_dict},
        }
        new_status = constants.ACTIVE
        placement_attr = vnf_dict['placement_attr']
        region_name = placement_attr.get('region_name')

        try:
            self._vnf_manager.invoke(
                driver_name, 'heal_wait', plugin=self,
                context=context, vnf_dict=vnf_dict, auth_attr=vim_auth,
                region_name=region_name)
            self.mgmt_call(context, vnf_dict, kwargs)

        except vnfm.VNFHealWaitFailed as e:
            with excutils.save_and_reraise_exception():
                new_status = constants.ERROR
                self._vnf_monitor.delete_hosting_vnf(vnf_dict['id'])
                self._vnf_maintenance_plugin.post(context, vnf_dict)
                self.set_vnf_error_status_reason(context, vnf_dict['id'],
                                                 str(e))
        except exceptions.MgmtDriverException as e:
            LOG.error('VNF configuration failed')
            new_status = constants.ERROR
            self._vnf_monitor.delete_hosting_vnf(vnf_dict['id'])
            self._vnf_maintenance_plugin.post(context, vnf_dict)
            self.set_vnf_error_status_reason(context, vnf_dict['id'], str(e))

        del vnf_dict['heal_stack_id']
        vnf_dict['status'] = new_status
        self.mgmt_update_post(context, vnf_dict)

        # Update vnf status to 'ACTIVE' so that monitoring can be resumed.
        evt_details = ("Ends the heal vnf request for VNF '%s'" %
                       vnf_dict['id'])
        self._vnf_monitor.update_hosting_vnf(vnf_dict, evt_details)
        self._vnf_maintenance_plugin.update_vnf_instances(self, context,
                                                          vnf_dict)
        self._vnf_maintenance_plugin.post(context, vnf_dict)
        # _update_vnf_post() method updates vnf_status and mgmt_ip_address
        self._update_vnf_post(context, vnf_dict['id'],
                              new_status, vnf_dict,
                              constants.PENDING_HEAL,
                              constants.RES_EVT_HEAL)

    def _update_vnf_wait(self, context, vnf_dict, vim_auth, driver_name):
        kwargs = {
            mgmt_constants.KEY_ACTION: mgmt_constants.ACTION_UPDATE_VNF,
            mgmt_constants.KEY_KWARGS: {'vnf': vnf_dict},
        }
        new_status = constants.ACTIVE
        placement_attr = vnf_dict['placement_attr']
        region_name = placement_attr.get('region_name')

        try:
            self._vnf_manager.invoke(
                driver_name, 'update_wait', plugin=self,
                context=context, vnf_dict=vnf_dict, auth_attr=vim_auth,
                region_name=region_name)

            self.mgmt_call(context, vnf_dict, kwargs)
        except vnfm.VNFUpdateWaitFailed as e:
            with excutils.save_and_reraise_exception():
                new_status = constants.ERROR
                self._vnf_monitor.delete_hosting_vnf(vnf_dict['id'])
                self.set_vnf_error_status_reason(context, vnf_dict['id'],
                                                 str(e))
        except exceptions.MgmtDriverException as e:
            LOG.error('VNF configuration failed')
            new_status = constants.ERROR
            self._vnf_monitor.delete_hosting_vnf(vnf_dict['id'])
            self.set_vnf_error_status_reason(context, vnf_dict['id'], str(e))
        vnf_dict['status'] = new_status
        self.mgmt_update_post(context, vnf_dict)
        self._update_vnf_post(context, vnf_dict['id'], new_status,
                              vnf_dict, constants.PENDING_UPDATE,
                              constants.RES_EVT_UPDATE)
        self._vnf_maintenance_plugin.create_vnf_constraints(self, context,
                                                            vnf_dict)

    def update_vnf(self, context, vnf_id, vnf):
        vnf_attributes = vnf['vnf']['attributes']
        if vnf_attributes.get('config'):
            config = vnf_attributes['config']
            if isinstance(config, dict):
                # TODO(sripriya) remove this yaml dump once db supports storing
                # json format of yaml files in a separate column instead of
                #  key value string pairs in vnf attributes table
                vnf_attributes['config'] = yaml.safe_dump(config)
            else:
                raise vnfm.InvalidAPIAttributeType(atype=type(config))

        if vnf_attributes.get('param_values'):
            param = vnf_attributes['param_values']
            if isinstance(param, dict):
                # TODO(sripriya) remove this yaml dump once db supports storing
                # json format of yaml files in a separate column instead of
                #  key value string pairs in vnf attributes table
                vnf_attributes['param_values'] = yaml.safe_dump(param)
            else:
                raise vnfm.InvalidAPIAttributeType(atype=type(param))

        vnf_dict = self._update_vnf_pre(context, vnf_id,
                                        constants.PENDING_UPDATE)
        driver_name, vim_auth = self._get_infra_driver(context, vnf_dict)
        instance_id = self._instance_id(vnf_dict)

        try:
            self.mgmt_update_pre(context, vnf_dict)
            self._vnf_manager.invoke(
                driver_name, 'update', plugin=self, context=context,
                vnf_id=instance_id, vnf_dict=vnf_dict,
                vnf=vnf, auth_attr=vim_auth)
        except vnfm.VNFUpdateInvalidInput:
            with excutils.save_and_reraise_exception():
                vnf_dict['status'] = constants.ACTIVE
                self._update_vnf_post(context, vnf_id,
                                      constants.ACTIVE,
                                      vnf_dict, constants.PENDING_UPDATE,
                                      constants.RES_EVT_UPDATE)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                vnf_dict['status'] = constants.ERROR
                self._vnf_monitor.delete_hosting_vnf(vnf_id)
                self.set_vnf_error_status_reason(context, vnf_dict['id'],
                                                 str(e))
                self.mgmt_update_post(context, vnf_dict)
                self._update_vnf_post(context, vnf_id,
                                      constants.ERROR,
                                      vnf_dict, constants.PENDING_UPDATE,
                                      constants.RES_EVT_UPDATE)

        self.spawn_n(self._update_vnf_wait, context, vnf_dict, vim_auth,
                     driver_name)
        return vnf_dict

    def heal_vnf(self, context, vnf_id, heal_request_data_obj):
        vnf_dict = self._update_vnf_pre(context, vnf_id,
                                        constants.PENDING_HEAL)
        driver_name, vim_auth = self._get_infra_driver(context, vnf_dict)
        # Update vnf status to 'PENDING_HEAL' so that monitoring can
        # be paused.
        evt_details = ("Starts heal vnf request for VNF '%s'. "
                       "Reason to Heal VNF: '%s'" % (vnf_dict['id'],
                       heal_request_data_obj.cause))
        self._vnf_monitor.update_hosting_vnf(vnf_dict, evt_details)

        try:
            vnf_dict['heal_stack_id'] = heal_request_data_obj.stack_id
            self._vnf_maintenance_plugin.project_instance_pre(context,
                                                              vnf_dict)
            self.mgmt_update_pre(context, vnf_dict)
            self._vnf_manager.invoke(
                driver_name, 'heal_vdu', plugin=self,
                context=context, vnf_dict=vnf_dict,
                heal_request_data_obj=heal_request_data_obj)
        except vnfm.VNFHealFailed as e:
            with excutils.save_and_reraise_exception():
                vnf_dict['status'] = constants.ERROR
                self._vnf_monitor.delete_hosting_vnf(vnf_id)
                self.set_vnf_error_status_reason(context, vnf_dict['id'],
                                                 str(e))
                self.mgmt_update_post(context, vnf_dict)
                self._vnf_maintenance_plugin.post(context, vnf_dict)
                self._update_vnf_post(context, vnf_id,
                                      constants.ERROR,
                                      vnf_dict, constants.PENDING_HEAL,
                                      constants.RES_EVT_HEAL)

        self.spawn_n(self._heal_vnf_wait, context, vnf_dict, vim_auth,
                     driver_name)

        return vnf_dict

    def _delete_vnf_wait(self, context, vnf_dict, auth_attr, driver_name):
        instance_id = self._instance_id(vnf_dict)
        e = None
        if instance_id:
            placement_attr = vnf_dict['placement_attr']
            region_name = placement_attr.get('region_name')
            try:
                self._vnf_manager.invoke(
                    driver_name,
                    'delete_wait',
                    plugin=self,
                    context=context,
                    vnf_id=instance_id,
                    auth_attr=auth_attr,
                    region_name=region_name)
            except Exception as e_:
                e = e_
                vnf_dict['status'] = constants.ERROR
                vnf_dict['error_reason'] = str(e)
                LOG.exception('_delete_vnf_wait')
                self.set_vnf_error_status_reason(context, vnf_dict['id'],
                                                 vnf_dict['error_reason'])

        self._vnf_maintenance_plugin.delete_vnf_constraints(self, context,
                                                            vnf_dict)
        self.mgmt_delete_post(context, vnf_dict)
        self._delete_vnf_post(context, vnf_dict, e)

    def _delete_vnf(self, context, vnf_id, force_delete=False):

        vnf_dict = self._delete_vnf_pre(context, vnf_id,
                                        force_delete=force_delete)
        driver_name, vim_auth = self._get_infra_driver(context, vnf_dict)
        self._vnf_monitor.delete_hosting_vnf(vnf_id)
        instance_id = self._instance_id(vnf_dict)
        placement_attr = vnf_dict['placement_attr']
        region_name = placement_attr.get('region_name')
        kwargs = {
            mgmt_constants.KEY_ACTION: mgmt_constants.ACTION_DELETE_VNF,
            mgmt_constants.KEY_KWARGS: {'vnf': vnf_dict},
        }
        try:
            self._vnf_maintenance_plugin.project_instance_pre(context,
                                                              vnf_dict)
            self.mgmt_delete_pre(context, vnf_dict)
            self.mgmt_call(context, vnf_dict, kwargs)
            if instance_id:
                self._vnf_manager.invoke(driver_name,
                                         'delete',
                                         plugin=self,
                                         context=context,
                                         vnf_id=instance_id,
                                         auth_attr=vim_auth,
                                         region_name=region_name)
        except Exception as e:
            # TODO(yamahata): when the device is already deleted. mask
            # the error, and delete row in db
            # Other case mark error
            with excutils.save_and_reraise_exception():
                if not force_delete:
                    vnf_dict['status'] = constants.ERROR
                    vnf_dict['error_reason'] = str(e)
                    self.set_vnf_error_status_reason(context, vnf_dict['id'],
                                                     vnf_dict['error_reason'])
                    self.mgmt_delete_post(context, vnf_dict)
                    self._delete_vnf_post(context, vnf_dict, e)

        if force_delete:
            self._delete_vnf_force(context, vnf_dict['id'])
            self.mgmt_delete_post(context, vnf_dict)
            self._delete_vnf_post(context, vnf_dict, None, force_delete=True)
        else:
            self.spawn_n(self._delete_vnf_wait, context, vnf_dict, vim_auth,
                         driver_name)

    def delete_vnf(self, context, vnf_id, vnf=None):

        # Extract "force_delete" from request's body
        force_delete = False
        if vnf and vnf['vnf'].get('attributes').get('force'):
            force_delete = vnf['vnf'].get('attributes').get('force')
        if force_delete and not context.is_admin:
            LOG.warning("force delete is admin only operation")
            raise exceptions.AdminRequired(reason="Admin only operation")

        self._delete_vnf(context, vnf_id, force_delete=force_delete)

    def _handle_vnf_scaling(self, context, policy):
        # validate
        def _validate_scaling_policy():
            type = policy['type']

            if type not in constants.POLICY_ACTIONS:
                raise exceptions.VnfPolicyTypeInvalid(
                    type=type,
                    valid_types=constants.POLICY_ACTIONS.keys(),
                    policy=policy['name']
                )
            action = policy['action']

            if action not in constants.POLICY_ACTIONS[type]:
                raise exceptions.VnfPolicyActionInvalid(
                    action=action,
                    valid_actions=constants.POLICY_ACTIONS[type],
                    policy=policy['name']
                )

            LOG.debug("Policy %s is validated successfully", policy['name'])

        def _get_status():
            if policy['action'] == constants.ACTION_SCALE_IN:
                status = constants.PENDING_SCALE_IN
            else:
                status = constants.PENDING_SCALE_OUT

            return status

        # pre
        def _handle_vnf_scaling_pre():
            status = _get_status()
            result = self._update_vnf_scaling_status(context,
                                                     policy,
                                                     [constants.ACTIVE],
                                                     status)
            LOG.debug("Policy %(policy)s vnf is at %(status)s",
                      {'policy': policy['name'],
                       'status': status})
            self._vnf_maintenance_plugin.project_instance_pre(context,
                                                              result)
            return result

        # post
        def _handle_vnf_scaling_post(new_status, mgmt_ip_address=None):
            status = _get_status()
            result = self._update_vnf_scaling_status(context,
                                                     policy,
                                                     [status],
                                                     new_status,
                                                     mgmt_ip_address)
            LOG.debug("Policy %(policy)s vnf is at %(status)s",
                      {'policy': policy['name'],
                       'status': new_status})
            action = 'delete' if policy['action'] == 'in' else 'update'
            self._vnf_maintenance_plugin.update_vnf_instances(self, context,
                                                              result,
                                                              action=action)
            return result

        # action
        def _vnf_policy_action():
            try:
                last_event_id = self._vnf_manager.invoke(
                    infra_driver,
                    'scale',
                    plugin=self,
                    context=context,
                    auth_attr=vim_auth,
                    policy=policy,
                    region_name=region_name
                )
                LOG.debug("Policy %s action is started successfully",
                          policy['name'])
                return last_event_id
            except Exception as e:
                LOG.error("Policy %s action is failed to start",
                          policy)
                with excutils.save_and_reraise_exception():
                    vnf['status'] = constants.ERROR
                    self.set_vnf_error_status_reason(
                        context, policy['vnf']['id'], str(e))
                    _handle_vnf_scaling_post(constants.ERROR)

        # wait
        def _vnf_policy_action_wait():
            try:
                LOG.debug("Policy %s action is in progress",
                          policy['name'])
                mgmt_ip_address = self._vnf_manager.invoke(
                    infra_driver,
                    'scale_wait',
                    plugin=self,
                    context=context,
                    auth_attr=vim_auth,
                    policy=policy,
                    region_name=region_name,
                    last_event_id=last_event_id
                )
                LOG.debug("Policy %s action is completed successfully",
                          policy['name'])
                _handle_vnf_scaling_post(constants.ACTIVE, mgmt_ip_address)
                # TODO(kanagaraj-manickam): Add support for config and mgmt
            except Exception as e:
                LOG.error("Policy %s action is failed to complete",
                          policy['name'])
                with excutils.save_and_reraise_exception():
                    self.set_vnf_error_status_reason(
                        context, policy['vnf']['id'], str(e))
                    _handle_vnf_scaling_post(constants.ERROR)

        _validate_scaling_policy()

        vnf = _handle_vnf_scaling_pre()
        policy['instance_id'] = vnf['instance_id']

        infra_driver, vim_auth = self._get_infra_driver(context, vnf)
        region_name = vnf.get('placement_attr', {}).get('region_name', None)
        last_event_id = _vnf_policy_action()
        self.spawn_n(_vnf_policy_action_wait)

        return policy

    def _make_policy_dict(self, vnf, name, policy):
        p = {}
        p['type'] = policy.get('type')
        p['properties'] = policy.get('properties') or policy.get(
            'triggers') or policy.get('reservation')
        p['vnf'] = vnf
        p['name'] = name
        p['id'] = uuidutils.generate_uuid()
        return p

    def get_vnf_policies(
            self, context, vnf_id, filters=None, fields=None):
        vnf = self.get_vnf(context, vnf_id)
        vnfd_tmpl = yaml.safe_load(vnf['vnfd']['attributes']['vnfd'])
        policy_list = []

        polices = vnfd_tmpl['topology_template'].get('policies', [])
        for policy_dict in polices:
            for name, policy in policy_dict.items():
                def _add(policy):
                    p = self._make_policy_dict(vnf, name, policy)
                    p['name'] = name
                    policy_list.append(p)

                # Check for filters
                if filters.get('name') or filters.get('type'):
                    if name == filters.get('name'):
                        _add(policy)
                        break
                    elif policy['type'] == filters.get('type'):
                        _add(policy)
                        break
                    else:
                        continue

                _add(policy)

        return policy_list

    def get_vnf_policy(
            self, context, policy_id, vnf_id, fields=None):
        policies = self.get_vnf_policies(context,
                                         vnf_id,
                                         filters={'name': policy_id})
        if policies:
            return policies[0]
        else:
            return None

    def create_vnf_scale(self, context, vnf_id, scale):
        policy_ = self.get_vnf_policy(context,
                                      scale['scale']['policy'],
                                      vnf_id)
        if not policy_:
            raise exceptions.VnfPolicyNotFound(policy=scale['scale']['policy'],
                                               vnf_id=vnf_id)
        policy_.update({'action': scale['scale']['type']})
        self._handle_vnf_scaling(context, policy_)

        return scale['scale']

    def get_vnf_policy_by_type(self, context, vnf_id, policy_type=None, fields=None):             # noqa
        policies = self.get_vnf_policies(context,
                                         vnf_id,
                                         filters={'type': policy_type})
        if policies:
            return policies[0]

        raise exceptions.VnfPolicyTypeInvalid(
            type=policy_type, policy=None,
            valid_types=constants.VALID_POLICY_TYPES)

    def _validate_alarming_policy(self, context, vnf_id, trigger):
        # validate alarm status

        # Trigger will contain only one action in trigger['trigger'] as it
        # filtered in _get_vnf_triggers().
        # Getting action from trigger to decide which process_alarm_for_vnf
        # method will be called.
        if list(trigger['trigger'])[0]\
                in constants.RESERVATION_POLICY_ACTIONS:
            if not self._vnf_reservation_monitor.process_alarm_for_vnf(
                    vnf_id, trigger):
                raise exceptions.AlarmUrlInvalid(vnf_id=vnf_id)
        else:
            if not self._vnf_alarm_monitor.process_alarm_for_vnf(
                    vnf_id, trigger):
                raise exceptions.AlarmUrlInvalid(vnf_id=vnf_id)

        # validate policy action. if action is composite, split it.
        # ex: respawn%notify
        action = trigger['action_name']
        action_list = action.split('%')
        pl_action_dict = dict()
        pl_action_dict['policy_actions'] = dict()
        pl_action_dict['policy_actions']['def_actions'] = list()
        pl_action_dict['policy_actions']['custom_actions'] = dict()
        for action in action_list:
            # validate policy action. if action is composite, split it.
            # ex: SP1-in, SP1-out
            action_ = None
            if action in constants.DEFAULT_ALARM_ACTIONS:
                pl_action_dict['policy_actions']['def_actions'].append(action)
            policy_ = self.get_vnf_policy(context, action, vnf_id)
            if not policy_:
                sp_action = action.split('-')
                if len(sp_action) == 2:
                    bk_policy_name = sp_action[0]
                    bk_policy_action = sp_action[1]
                    policies_ = self.get_vnf_policies(
                        context, vnf_id, filters={'name': bk_policy_name})
                    if policies_:
                        policy_ = policies_[0]
                        action_ = bk_policy_action
            if policy_:
                pl_action_dict['policy_actions']['custom_actions'].update(
                    {policy_['id']: {'bckend_policy': policy_,
                                   'bckend_action': action_}})

            LOG.debug("Trigger %s is validated successfully", trigger)

        return pl_action_dict
        # validate url

    def _get_vnf_triggers(self, context, vnf_id, filters=None, fields=None):
        if filters.get('name') in constants.RESERVATION_POLICY_ACTIONS:
            policy = self.get_vnf_policy_by_type(
                context, vnf_id, policy_type=constants.POLICY_RESERVATION)
        else:
            policy = self.get_vnf_policy_by_type(
                context, vnf_id, policy_type=constants.POLICY_ALARMING)
        triggers = policy['properties']
        vnf_trigger = dict()
        for trigger_name, trigger_dict in triggers.items():
            if trigger_name == filters.get('name'):
                vnf_trigger['trigger'] = {trigger_name: trigger_dict}
                vnf_trigger['vnf'] = policy['vnf']
                break

        return vnf_trigger

    def get_vnf_trigger(self, context, vnf_id, trigger_name):
        trigger = self._get_vnf_triggers(
            context, vnf_id, filters={'name': trigger_name})
        if not trigger:
            raise exceptions.TriggerNotFound(
                trigger_name=trigger_name,
                vnf_id=vnf_id
            )
        return trigger

    def _handle_vnf_monitoring(self, context, trigger):

        vnf_dict = trigger['vnf']
        # Multiple actions support
        if trigger.get('policy_actions'):
            policy_actions = trigger['policy_actions']
            if policy_actions.get('def_actions'):
                for action in policy_actions['def_actions']:
                    self._vnf_action.invoke(
                        action, 'execute_action', plugin=self, context=context,
                        vnf_dict=vnf_dict, args={})
            if policy_actions.get('custom_actions'):
                custom_actions = policy_actions['custom_actions']
                for pl_action, pl_action_dict in custom_actions.items():
                    bckend_policy = pl_action_dict['bckend_policy']
                    bckend_action = pl_action_dict['bckend_action']
                    bckend_policy_type = bckend_policy['type']
                    if bckend_policy_type == constants.POLICY_SCALING:
                        if vnf_dict['status'] != constants.ACTIVE:
                            LOG.info(_("Scaling Policy action "
                                       "skipped due to status "
                                       "%(status)s for vnf %(vnfid)s") %
                                     {"status": vnf_dict['status'],
                                      "vnfid": vnf_dict['id']})
                            return
                        action = 'autoscaling'
                        scale = {}
                        scale.setdefault('scale', {})
                        scale['scale']['type'] = bckend_action
                        scale['scale']['policy'] = bckend_policy['name']
                        self._vnf_action.invoke(
                            action, 'execute_action', plugin=self,
                            context=context, vnf_dict=vnf_dict, args=scale)

    def create_vnf_trigger(
            self, context, vnf_id, trigger):
        trigger_ = self.get_vnf_trigger(
            context, vnf_id, trigger['trigger']['policy_name'])
        # action_name before analyzing
        trigger_.update({'action_name': trigger['trigger']['action_name']})
        trigger_.update({'params': trigger['trigger']['params']})
        policy_actions = self._validate_alarming_policy(
            context, vnf_id, trigger_)
        if policy_actions:
            trigger_.update(policy_actions)
        self._handle_vnf_monitoring(context, trigger_)
        return trigger['trigger']

    def get_vnf_resources(self, context, vnf_id, fields=None, filters=None):
        vnf_info = self.get_vnf(context, vnf_id)
        infra_driver, vim_auth = self._get_infra_driver(context, vnf_info)
        region_name = vnf_info.get('placement_attr', {}).\
            get('region_name', None)
        if vnf_info['status'] == constants.ACTIVE:
            vnf_details = self._vnf_manager.invoke(infra_driver,
                                                   'get_resource_info',
                                                   plugin=self,
                                                   context=context,
                                                   vnf_info=vnf_info,
                                                   auth_attr=vim_auth,
                                                   region_name=region_name)
            resources = [{'name': name,
                          'type': info.get('type'),
                          'id': info.get('id')}
                        for name, info in vnf_details.items()]
            return resources
        # Raise exception when VNF.status != ACTIVE
        else:
            raise vnfm.VNFInactive(vnf_id=vnf_id,
                                   message=_(' Cannot fetch details'))

    def create_vnf_maintenance(self, context, vnf_id, maintenance):
        _maintenance = self._vnf_maintenance_plugin.validate_maintenance(
            maintenance.copy())
        vnf = self.get_vnf(context, vnf_id)
        _maintenance['vnf'] = vnf
        self._vnf_maintenance_plugin.handle_maintenance(
            self, context, _maintenance)
        policy_action = _maintenance.get('policy_action', '')
        if policy_action:
            self._vnf_action.invoke(
                policy_action['action'], 'execute_action', plugin=self,
                context=context, vnf_dict=vnf, args=policy_action['args'])
        else:
            self._vnf_maintenance_plugin.request(self, context, vnf,
                                                 _maintenance)
        return maintenance['maintenance']



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker.objects import base
from tacker.objects import fields

LOG = logging.getLogger(__name__)


@base.TackerObjectRegistry.register
class CancelMode(base.TackerObject, base.TackerPersistentObject):
    # SOL 003 5.5.2.14 Type: CancelMode
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'cancel_mode': fields.VnfInstanceTerminationTypeField(
            nullable=False),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            return super(
                CancelMode, cls).obj_from_primitive(primitive, context)
        return CancelMode._from_dict(primitive)

    @classmethod
    def _from_dict(cls, data_dict):
        return cls(cancel_mode=data_dict.get('cancel_mode'))



# Copyright 2015 Brocade Communications System, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from oslo_config import cfg

from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file

CONF = cfg.CONF
VNF_CIRROS_CREATE_TIMEOUT = 120


class VnfTestCreate(base.BaseTackerTest):
    def _test_create_delete_vnf(self, vnf_name, vnfd_name, vim_id=None):
        input_yaml = read_file('sample-tosca-vnfd-no-monitor.yaml')
        tosca_dict = yaml.safe_load(input_yaml)
        tosca_arg = {'vnfd': {'name': vnfd_name,
                     'attributes': {'vnfd': tosca_dict}}}

        # Create vnfd with tosca template
        vnfd_instance = self.client.create_vnfd(body=tosca_arg)
        self.assertIsNotNone(vnfd_instance)

        # Create vnf with vnfd_id
        vnfd_id = vnfd_instance['vnfd']['id']
        self.addCleanup(self.client.delete_vnfd, vnfd_id)

        vnf_arg = {'vnf': {'vnfd_id': vnfd_id, 'name': vnf_name}}
        if vim_id:
            vnf_arg['vnf']['vim_id'] = vim_id
        vnf_instance = self.client.create_vnf(body=vnf_arg)
        self.validate_vnf_instance(vnfd_instance, vnf_instance)

        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        self.assertIsNotNone(self.client.show_vnf(vnf_id)['vnf'][
            'mgmt_ip_address'])
        if vim_id:
            self.assertEqual(vim_id, vnf_instance['vnf']['vim_id'])

        # Get vnf details when vnf is in active state
        vnf_details = self.client.list_vnf_resources(vnf_id)['resources'][0]
        self.assertIn('name', vnf_details)
        self.assertIn('id', vnf_details)
        self.assertIn('type', vnf_details)

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE,
            evt_constants.PENDING_CREATE, cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.ACTIVE)

        # update VIM name when VNFs are active.
        # check for exception.
        vim0_id = vnf_instance['vnf']['vim_id']
        msg = "VIM %s is still in use by VNF" % vim0_id
        try:
            update_arg = {'vim': {'name': "vnf_vim"}}
            self.client.update_vim(vim0_id, update_arg)
        except Exception as err:
            self.assertEqual(err.message, msg)
        else:
            self.assertTrue(
                False,
                "Name of vim(s) with active vnf(s) should not be changed!")

        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete failed"

        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)

    def test_create_delete_vnf_with_default_vim(self):
        self._test_create_delete_vnf(
            vnf_name='test_vnf_with_cirros_no_monitoring_default_vim',
            vnfd_name='sample_cirros_vnf_no_monitoring_default_vim')

    def test_create_delete_vnf_with_vim_id(self):
        vim_list = self.client.list_vims()
        vim0_id = self.get_vim(vim_list, 'VIM0')['id']
        self._test_create_delete_vnf(
            vim_id=vim0_id,
            vnf_name='test_vnf_with_cirros_vim_id',
            vnfd_name='sample_cirros_vnf_no_monitoring_vim_id')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.api import extensions


class VnffgAbstractDriver(extensions.PluginInterface, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_type(self):
        """Return one of predefined type of Tacker drivers."""
        pass

    @abc.abstractmethod
    def get_name(self):
        """Return a symbolic name for the Tacker VNFFG SFC driver."""
        pass

    @abc.abstractmethod
    def get_description(self):
        pass

    @abc.abstractmethod
    def create_chain(self, name, fc_id, vnfs, symmetrical=False,
                     auth_attr=None):
        """Create service function chain and returns an ID"""
        pass

    @abc.abstractmethod
    def update_chain(self, chain_id, fc_ids, vnfs,
                     symmetrical=False,
                     auth_attr=None):
        """Update service function chain"""
        pass

    @abc.abstractmethod
    def delete_chain(self, chain_id, auth_attr=None):
        """Delete service function chain"""
        pass

    @abc.abstractmethod
    def create_flow_classifier(self, name, fc, auth_attr=None):
        """Create flow classifier and returns an ID"""
        pass

    @abc.abstractmethod
    def update_flow_classifier(self, fc_id, fc, auth_attr=None):
        """Update flow classifier"""
        pass

    @abc.abstractmethod
    def delete_flow_classifier(self, fc_id, auth_attr=None):
        """Delete flow classifier"""
        pass



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re
import unittest

from tacker.objects import fields
from tacker.tests.functional.sol_kubernetes.vnflcm import base as vnflcm_base
from tacker.tests import utils


class VnfLcmKubernetesTest(vnflcm_base.BaseVnfLcmKubernetesTest):

    @classmethod
    def setUpClass(cls):
        super(VnfLcmKubernetesTest, cls).setUpClass()
        vnf_package_id, cls.vnfd_id = cls._create_and_upload_vnf_package(
            cls, cls.tacker_client, "test_cnf",
            {"key": "resource_functional_common"})
        cls.vnf_package_ids.append(vnf_package_id)

    @classmethod
    def tearDownClass(cls):
        super(VnfLcmKubernetesTest, cls).tearDownClass()

    @classmethod
    def _update_source_path(cls, meta_dir, meta_name, port):
        meta_path = os.path.join(meta_dir, meta_name)
        with open(file=meta_path, mode='r', encoding='UTF8') as f:
            meta_content = f.read()
        new_meta_content = re.sub(
            r':(\d{5})', ':' + str(port), meta_content)
        with open(meta_path, 'w', encoding='utf-8') as f:
            f.write(new_meta_content)

    def _test_inst_term_from_manifest(self, inst_name, inst_desc,
                                      additional_param, flavour_id="simple",
                                      vnfd_id=None):
        # Create and instantiate vnf instance
        if vnfd_id is None:
            vnfd_id = self.vnfd_id
        vnf_instance = self._create_and_instantiate_vnf_instance(
            vnfd_id, flavour_id, inst_name, inst_desc, additional_param)
        self.assertEqual(vnf_instance['instantiationState'], 'INSTANTIATED')

        # Terminate vnf forcefully
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.FORCEFUL,
        }
        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)
        self._delete_vnf_instance(vnf_instance['id'])

    # The same problem as
    # https://github.com/kubernetes-client/python/issues/547,
    # after fixing this bug, the bindings test items can pass normally.
    # def test_inst_term_cnf_with_binding(self):
    #     vnf_instance_name = "cnf_with_binding"
    #     vnf_instance_description = "cnf with binding"
    #     files = ["Files/kubernetes/bindings.yaml"]
    #     self._test_inst_term_from_manifest(
    #         vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_clusterrole_clusterrolebinding_SA(self):
        vnf_instance_name = "cnf_with_clusterrole_clusterrolebinding_SA"
        vnf_instance_description = "cnf with clusterRole/clusterRoleBinding/SA"
        files = ["Files/kubernetes/clusterrole_clusterrolebinding_SA.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_config_map(self):
        vnf_instance_name = "cnf_with_configMap"
        vnf_instance_description = "cnf with configMap"
        files = ["Files/kubernetes/config-map.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    @unittest.skip("Until BUG 1910327")
    def test_inst_term_cnf_with_controller_revision(self):
        vnf_instance_name = "cnf_with_controller_revision"
        vnf_instance_description = "cnf with controllerRevision"
        files = ["Files/kubernetes/controller-revision.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_daemon_set(self):
        vnf_instance_name = "cnf_with_daemon_set"
        vnf_instance_description = "cnf with daemonSet"
        files = ["Files/kubernetes/daemon-set.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_deployment(self):
        vnf_instance_name = "cnf_with_deployment"
        vnf_instance_description = "cnf with deployment"
        files = ["Files/kubernetes/deployment.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_horizontal_pod_autoscaler(self):
        vnf_instance_name = "cnf_with_horizontal_pod_autoscaler"
        vnf_instance_description = "cnf with horizontalPodAutoscaler"
        files = ["Files/kubernetes/horizontal-pod-autoscaler.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_job(self):
        vnf_instance_name = "cnf_with_job"
        vnf_instance_description = "cnf with job"
        files = ["Files/kubernetes/job.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_limit_range(self):
        vnf_instance_name = "cnf_with_limit_range"
        vnf_instance_description = "cnf with limitRange"
        files = ["Files/kubernetes/limit-range.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_namespace(self):
        vnf_instance_name = "cnf_with_namespae"
        vnf_instance_description = "cnf with namespace"
        files = ["Files/kubernetes/namespace.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_pod(self):
        vnf_instance_name = "cnf_with_pod"
        vnf_instance_description = "cnf with pod"
        files = ["Files/kubernetes/pod.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_pod_template(self):
        vnf_instance_name = "cnf_with_pod_template"
        vnf_instance_description = "cnf with podTemplate"
        files = ["Files/kubernetes/pod-template.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_resource_quota(self):
        vnf_instance_name = "cnf_with_resource_quota"
        vnf_instance_description = "cnf with resourceQuota"
        files = ["Files/kubernetes/resource-quota.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_role_rolebinding_SA(self):
        vnf_instance_name = "cnf_with_role_rolebinding_SA"
        vnf_instance_description = "cnf with role/roleBinding/SA"
        files = ["Files/kubernetes/role_rolebinding_SA.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_subject_access_review(self):
        vnf_instance_name = "cnf_with_subject_access_review"
        vnf_instance_description = "cnf with subjectAccessReview"
        files = ["Files/kubernetes/subject-access-review.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_token_review(self):
        vnf_instance_name = "cnf_with_token_review"
        vnf_instance_description = "cnf with tokenReview"
        files = ["Files/kubernetes/token-review.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_ext_artifact(self):
        # Setup http server
        instance_file_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '../../../etc/samples/etsi/nfv/test_cnf_ext_artifact')
        artifact_file_dir = os.path.join(
            instance_file_dir, 'Files/kubernetes')
        http_handler = utils.StaticHttpFileHandler(artifact_file_dir)
        self.addCleanup(http_handler.stop)
        artifact_file_url = (f"http://127.0.0.1:{http_handler.port}/"
                             "storage-class-url.yaml")
        meta_dir = os.path.join(instance_file_dir, 'TOSCA-Metadata')
        self._update_source_path(meta_dir, 'TOSCA.meta', http_handler.port)

        # Create and update vnf package
        vnf_package_id, vnfd_id = self._create_and_upload_vnf_package(
            self.tacker_client, "test_cnf_ext_artifact",
            {"key": "resource_functional_external_artifact"})
        self.vnf_package_ids.append(vnf_package_id)

        vnf_instance_name = "cnf_with_ext_artifact"
        vnf_instance_description = "cnf with ext artifact"
        files = [artifact_file_url]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param,
            vnfd_id=vnfd_id)

    def test_inst_term_cnf_in_multiple_yaml_with_single_resource(self):
        vnf_instance_name = "cnf_in_multiple_yaml_with_single_resource"
        vnf_instance_description = "cnf in multiple yaml with single resource"
        files = ["Files/kubernetes/replicaset_service_secret.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_in_single_yaml_with_multiple_resources(self):
        vnf_instance_name = "cnf_in_single_yaml_with_multiple_resources"
        vnf_instance_description = "cnf in single yaml with multiple resources"
        files = ["Files/kubernetes/multiple_yaml_priority-class.yaml",
                 "Files/kubernetes/multiple_yaml_lease.yaml",
                 "Files/kubernetes/multiple_yaml_network-policy.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_multi_yaml_and_resources_no_dep(self):
        vnf_instance_name = "cnf_multi_yaml_and_resources_no_dep"
        vnf_instance_description = "cnf multi yaml and resources no dep"
        files = ["Files/kubernetes/local-subject-access-review.yaml",
                 "Files/kubernetes/self-subject-access-review_"
                 "and_self-subject-rule-review.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_inst_term_cnf_with_multi_yaml_and_resources_dep_and_sort(self):
        vnf_instance_name = "cnf_multi_yaml_and_resources_dep_and_sort"
        vnf_instance_description = "cnf multi yaml and resources dep and sort"
        files = ["Files/kubernetes/storage-class.yaml",
                 "Files/kubernetes/persistent-volume-0.yaml",
                 "Files/kubernetes/persistent-volume-1.yaml",
                 "Files/kubernetes/statefulset.yaml",
                 "Files/kubernetes/storage-class_pv_pvc.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files}
        self._test_inst_term_from_manifest(
            vnf_instance_name, vnf_instance_description, additional_param)

    def test_rollback_cnf_after_instantiate_fail(self):
        vnf_instance_name = "vnf_rollback_cnf_after_instantiate_fail"
        vnf_instance_description = "vnf rollback cnf after instantiate fail"
        resp, vnf_instance = self._create_vnf_instance(
            self.vnfd_id,
            vnf_instance_name=vnf_instance_name,
            vnf_instance_description=vnf_instance_description)
        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)

        additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/statefulset_fail.yaml",
            ]
        }
        request_body = self._instantiate_vnf_instance_request(
            "simple", vim_id=self.vim_id, additional_param=additional_param)
        self._instantiate_vnf_instance(
            vnf_instance['id'], request_body, wait_state="FAILED_TEMP")
        self._test_rollback_cnf_instantiate(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])



# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from alembic import op
import contextlib
import sqlalchemy as sa
from sqlalchemy.engine import reflection


def alter_enum(table, column, enum_type, nullable):
    bind = op.get_bind()
    engine = bind.engine
    if engine.name == 'postgresql':
        values = {'table': table,
                  'column': column,
                  'name': enum_type.name}
        op.execute("ALTER TYPE %(name)s RENAME TO old_%(name)s" % values)
        enum_type.create(bind, checkfirst=False)
        op.execute("ALTER TABLE %(table)s RENAME COLUMN %(column)s TO "
                   "old_%(column)s" % values)
        op.add_column(table, sa.Column(column, enum_type, nullable=nullable))
        op.execute("UPDATE %(table)s SET %(column)s = "
                   "old_%(column)s::text::%(name)s" % values)
        op.execute("ALTER TABLE %(table)s DROP COLUMN old_%(column)s" % values)
        op.execute("DROP TYPE old_%(name)s" % values)
    else:
        op.alter_column(table, column, type_=enum_type,
                        existing_nullable=nullable)


def create_foreign_key_constraint(table_name, fk_constraints):
    for fk in fk_constraints:
        op.create_foreign_key(
            constraint_name=fk['name'],
            source_table=table_name,
            referent_table=fk['referred_table'],
            local_cols=fk['constrained_columns'],
            remote_cols=fk['referred_columns'],
            ondelete=fk['options'].get('ondelete')
        )


def drop_foreign_key_constraint(table_name, fk_constraints):
    for fk in fk_constraints:
        op.drop_constraint(
            constraint_name=fk['name'],
            table_name=table_name,
            type_='foreignkey'
        )


@contextlib.contextmanager
def modify_foreign_keys_constraint(table_names):
    inspector = reflection.Inspector.from_engine(op.get_bind())
    try:
        for table in table_names:
            fk_constraints = inspector.get_foreign_keys(table)
            drop_foreign_key_constraint(table, fk_constraints)
        yield
    finally:
        for table in table_names:
            fk_constraints = inspector.get_foreign_keys(table)
            create_foreign_key_constraint(table, fk_constraints)


def modify_foreign_keys_constraint_with_col_change(
        table_name, old_local_col, new_local_col, existing_type,
        nullable=False):
    inspector = reflection.Inspector.from_engine(op.get_bind())
    fk_constraints = inspector.get_foreign_keys(table_name)
    for fk in fk_constraints:
        if old_local_col in fk['constrained_columns']:
            drop_foreign_key_constraint(table_name, [fk])
    op.alter_column(table_name, old_local_col,
                    new_column_name=new_local_col,
                    existing_type=existing_type,
                    nullable=nullable)
    fk_constraints = inspector.get_foreign_keys(table_name)
    for fk in fk_constraints:
        for i in range(len(fk['constrained_columns'])):
            if old_local_col == fk['constrained_columns'][i]:
                fk['constrained_columns'][i] = new_local_col
                create_foreign_key_constraint(table_name, [fk])
                break



# Copyright 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_name_to_classifiers

Revision ID: 5d490546290c
Revises: e9a1e47fb0b5
Create Date: 2018-01-11 14:27:16.334946

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '5d490546290c'
down_revision = 'e9a1e47fb0b5'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.add_column('vnffgclassifiers',
                  sa.Column('name', sa.String(255),
                            nullable=True, server_default=''))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker import objects
from tacker.objects import base
from tacker.objects import fields

LOG = logging.getLogger(__name__)


@base.TackerObjectRegistry.register
class ChangeExtConnRequest(base.TackerObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vim_connection_info': fields.ListOfObjectsField(
            'VimConnectionInfo', nullable=True, default=[]),
        'ext_virtual_links': fields.ListOfObjectsField(
            'ExtVirtualLinkData', nullable=False),
        'additional_params': fields.DictOfNullableField(nullable=True,
            default={})
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_change_ext_conn_req = super(
                ChangeExtConnRequest, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'vim_connection_info' in primitive.keys():
                obj_data = [objects.VimConnectionInfo._from_dict(
                    vim_conn) for vim_conn in primitive.get(
                    'vim_connection_info', [])]
                primitive.update({'vim_connection_info': obj_data})

            if 'ext_virtual_links' in primitive.keys():
                obj_data = [objects.ExtVirtualLinkData.obj_from_primitive(
                    ext_vir_link, context) for ext_vir_link in primitive.get(
                    'ext_virtual_links', [])]
                primitive.update({'ext_virtual_links': obj_data})
            obj_change_ext_conn_req = ChangeExtConnRequest._from_dict(
                primitive)

        return obj_change_ext_conn_req

    @classmethod
    def _from_dict(cls, data_dict):
        vim_connection_info = data_dict.get('vim_connection_info', [])
        ext_virtual_links = data_dict.get('ext_virtual_links', [])
        additional_params = data_dict.get('additional_params', {})

        return cls(
            vim_connection_info=vim_connection_info,
            ext_virtual_links=ext_virtual_links,
            additional_params=additional_params)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""multisite_vim

Revision ID: 5246a6bd410f
Revises: 24bec5f211c7
Create Date: 2016-03-22 14:05:15.129330

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '5246a6bd410f'
down_revision = '24bec5f211c7'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.create_table('vims',
        sa.Column('id', sa.String(length=255), nullable=False),
        sa.Column('type', sa.String(length=255), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('placement_attr', sa.PickleType(), nullable=True),
        sa.Column('shared', sa.Boolean(), server_default=sa.text('true'),
                  nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_table('vimauths',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('vim_id', sa.String(length=255), nullable=False),
        sa.Column('password', sa.String(length=128), nullable=False),
        sa.Column('auth_url', sa.String(length=255), nullable=False),
        sa.Column('vim_project', sa.PickleType(), nullable=False),
        sa.Column('auth_cred', sa.PickleType(), nullable=False),
        sa.ForeignKeyConstraint(['vim_id'], ['vims.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('auth_url')
    )
    op.add_column('devices', sa.Column('placement_attr', sa.PickleType(),
                                        nullable=True))
    op.add_column('devices', sa.Column('vim_id', sa.String(length=36),
                                        nullable=False))
    op.create_foreign_key(None, 'devices', 'vims', ['vim_id'], ['id'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_config import cfg
from toscaparser import tosca_template
import unittest
import yaml


from tacker.common import utils
from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file
from tacker.tosca import utils as toscautils

CONF = cfg.CONF
SOFTWARE_DEPLOYMENT = 'OS::Heat::SoftwareDeployment'


class VnfTestToscaVNFC(base.BaseTackerTest):

    @unittest.skip("Until BUG 1673012")
    def test_create_delete_tosca_vnfc(self):
        input_yaml = read_file('sample_tosca_vnfc.yaml')
        tosca_dict = yaml.safe_load(input_yaml)
        path = os.path.abspath(os.path.join(
            os.path.dirname(__file__), "../../../etc/samples"))
        vnfd_name = 'sample-tosca-vnfc'
        tosca_dict['topology_template']['node_templates'
                                        ]['firewall_vnfc'
                                          ]['interfaces'
                                            ]['Standard']['create'] = path \
            + '/install_vnfc.sh'
        tosca_arg = {'vnfd': {'name': vnfd_name,
                              'attributes': {'vnfd': tosca_dict}}}

        # Create vnfd with tosca template
        vnfd_instance = self.client.create_vnfd(body=tosca_arg)
        self.assertIsNotNone(vnfd_instance)

        # Create vnf with vnfd_id
        vnfd_id = vnfd_instance['vnfd']['id']
        self.addCleanup(self.client.delete_vnfd, vnfd_id)
        vnf_arg = {'vnf': {'vnfd_id': vnfd_id, 'name':
                           "test_tosca_vnfc"}}
        vnf_instance = self.client.create_vnf(body=vnf_arg)

        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(vnf_id,
                                   constants.VNFC_CREATE_TIMEOUT,
                                   constants.ACTIVE_SLEEP_TIME)
        self.assertEqual('ACTIVE',
                         self.client.show_vnf(vnf_id)['vnf']['status'])
        self.validate_vnf_instance(vnfd_instance, vnf_instance)

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.PENDING_CREATE,
            cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.ACTIVE)

        # Validate mgmt_ip_address with input yaml file
        mgmt_ip_address = self.client.show_vnf(vnf_id)['vnf'][
            'mgmt_ip_address']
        self.assertIsNotNone(mgmt_ip_address)
        mgmt_dict = yaml.safe_load(str(mgmt_ip_address))

        input_dict = yaml.safe_load(input_yaml)
        toscautils.updateimports(input_dict)

        tosca = tosca_template.ToscaTemplate(parsed_params={}, a_file=False,
                          yaml_dict_tpl=input_dict)

        vdus = toscautils.findvdus(tosca)

        self.assertEqual(len(vdus), len(mgmt_dict.keys()))
        for vdu in vdus:
            self.assertIsNotNone(mgmt_dict[vdu.name])
            self.assertEqual(True, utils.is_valid_ipv4(mgmt_dict[vdu.name]))

        # Check the status of SoftwareDeployment
        heat_stack_id = self.client.show_vnf(vnf_id)['vnf']['instance_id']
        resource_types = self.h_client.resources
        resources = resource_types.list(stack_id=heat_stack_id)
        for resource in resources:
            resource = resource.to_dict()
            if resource['resource_type'] == \
                    SOFTWARE_DEPLOYMENT:
                self.assertEqual('CREATE_COMPLETE',
                    resource['resource_status'])
                break

        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete of test_vnf_with_multiple_vdus failed"

        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from datetime import datetime

from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker.sol_refactored.api.api_version import supported_versions_v2
from tacker.sol_refactored.api.schemas import vnflcm_v2 as schema
from tacker.sol_refactored.api import validator
from tacker.sol_refactored.api import wsgi as sol_wsgi
from tacker.sol_refactored.common import config
from tacker.sol_refactored.common import coordinate
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import lcm_op_occ_utils as lcmocc_utils
from tacker.sol_refactored.common import subscription_utils as subsc_utils
from tacker.sol_refactored.common import vim_utils
from tacker.sol_refactored.common import vnf_instance_utils as inst_utils
from tacker.sol_refactored.conductor import conductor_rpc_v2
from tacker.sol_refactored.controller import vnflcm_view
from tacker.sol_refactored.nfvo import nfvo_client
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields as v2fields


LOG = logging.getLogger(__name__)  # NOTE: unused at the moment

CONF = config.CONF


class VnfLcmControllerV2(sol_wsgi.SolAPIController):

    def __init__(self):
        self.nfvo_client = nfvo_client.NfvoClient()
        self.endpoint = CONF.v2_vnfm.endpoint
        self.conductor_rpc = conductor_rpc_v2.VnfLcmRpcApiV2()
        self._inst_view = vnflcm_view.InstanceViewBuilder(self.endpoint)
        self._lcmocc_view = vnflcm_view.LcmOpOccViewBuilder(self.endpoint)
        self._subsc_view = vnflcm_view.SubscriptionViewBuilder(self.endpoint)

    def api_versions(self, request):
        return sol_wsgi.SolResponse(200, supported_versions_v2)

    @validator.schema(schema.CreateVnfRequest_V200, '2.0.0')
    def create(self, request, body):
        context = request.context
        vnfd_id = body['vnfdId']

        pkg_info = self.nfvo_client.get_vnf_package_info_vnfd(
            context, vnfd_id)
        if pkg_info.operationalState != "ENABLED":
            raise sol_ex.VnfdIdNotEnabled(vnfd_id=vnfd_id)

        vnfd = self.nfvo_client.get_vnfd(context, vnfd_id)
        vnfd_prop = vnfd.get_vnfd_properties()

        metadata = vnfd_prop['metadata']
        if 'metadata' in body:
            metadata = inst_utils.json_merge_patch(metadata, body['metadata'])

        inst = objects.VnfInstanceV2(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfdId=vnfd_id,
            vnfProvider=pkg_info.vnfProvider,
            vnfProductName=pkg_info.vnfProductName,
            vnfSoftwareVersion=pkg_info.vnfSoftwareVersion,
            vnfdVersion=pkg_info.vnfdVersion,
            instantiationState='NOT_INSTANTIATED',
            # not set at the moment. will be set when instantiate.
            # vimConnectionInfo
            # instantiatedVnfInfo
        )

        # set optional fields
        if body.get('vnfInstanceName'):
            inst.vnfInstanceName = body['vnfInstanceName']
        if body.get('vnfInstanceDescription'):
            inst.vnfInstanceDescription = body['vnfInstanceDescription']
        if vnfd_prop.get('vnfConfigurableProperties'):
            inst.vnfConfigurableProperties = vnfd_prop[
                'vnfConfigurableProperties']
        if metadata:
            inst.metadata = metadata
        if vnfd_prop.get('extensions'):
            inst.extensions = vnfd_prop['extensions']

        inst.create(context)

        self.nfvo_client.send_inst_create_notification(context, inst,
                                                       self.endpoint)
        resp_body = self._inst_view.detail(inst)
        location = inst_utils.inst_href(inst.id, self.endpoint)

        return sol_wsgi.SolResponse(201, resp_body, location=location)

    def index(self, request):
        filter_param = request.GET.get('filter')
        if filter_param is not None:
            filters = self._inst_view.parse_filter(filter_param)
        else:
            filters = None
        # validate_filter

        selector = self._inst_view.parse_selector(request.GET)

        insts = inst_utils.get_inst_all(request.context)

        resp_body = self._inst_view.detail_list(insts, filters, selector)

        return sol_wsgi.SolResponse(200, resp_body)

    def show(self, request, id):
        inst = inst_utils.get_inst(request.context, id)

        resp_body = self._inst_view.detail(inst)

        return sol_wsgi.SolResponse(200, resp_body)

    @coordinate.lock_vnf_instance('{id}')
    def delete(self, request, id):
        context = request.context
        inst = inst_utils.get_inst(context, id)

        if inst.instantiationState != 'NOT_INSTANTIATED':
            raise sol_ex.VnfInstanceIsInstantiated(inst_id=id)

        lcmocc_utils.check_lcmocc_in_progress(context, id)

        inst.delete(context)

        # NOTE: inst record in DB deleted but inst object still
        # can be accessed.
        self.nfvo_client.send_inst_delete_notification(context, inst,
                                                       self.endpoint)
        return sol_wsgi.SolResponse(204, None)

    def _new_lcmocc(self, inst_id, operation, req_body,
            op_state=v2fields.LcmOperationStateType.STARTING):
        now = datetime.utcnow()
        lcmocc = objects.VnfLcmOpOccV2(
            id=uuidutils.generate_uuid(),
            operationState=op_state,
            stateEnteredTime=now,
            startTime=now,
            vnfInstanceId=inst_id,
            operation=operation,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req_body)

        return lcmocc

    @validator.schema(schema.VnfInfoModificationRequest_V200, '2.0.0')
    @coordinate.lock_vnf_instance('{id}')
    def update(self, request, id, body):
        context = request.context
        inst = inst_utils.get_inst(context, id)

        lcmocc_utils.check_lcmocc_in_progress(context, id)
        if (inst.instantiationState == 'NOT_INSTANTIATED'
                and 'vimConnectionInfo' in body):
            msg = 'vimConnectionInfo cannot be modified in NOT_INSTANTIATED.'
            raise sol_ex.SolValidationError(detail=msg)

        # check vnf package operational state
        if 'vnfdId' in body and body['vnfdId'] != inst.vnfdId:
            req_vnfd_id = body['vnfdId']
            pkg_info = self.nfvo_client.get_vnf_package_info_vnfd(
                context, req_vnfd_id)
            if pkg_info.operationalState != "ENABLED":
                # NOTE: Response code 422 is not specified in SOL003,
                # but 422 will be returned here. This is because
                # instance create returns 422 for the same reason.
                # In addition, SOL013 defines 422 as common error situations.
                raise sol_ex.VnfdIdNotEnabled(vnfd_id=req_vnfd_id)

        if 'vnfcInfoModifications' in body:
            vnfc_id = []
            if inst.obj_attr_is_set('instantiatedVnfInfo'):
                inst_info = inst.instantiatedVnfInfo
                if inst_info.obj_attr_is_set('vnfcInfo'):
                    vnfc_id = [vnfc.id for vnfc in inst_info.vnfcInfo]
            for vnfc_mod in body['vnfcInfoModifications']:
                if vnfc_mod['id'] not in vnfc_id:
                    raise sol_ex.SolValidationError(
                        detail="vnfcInstanceId(%s) does not exist."
                        % vnfc_mod['id'])

        lcmocc = self._new_lcmocc(id, v2fields.LcmOperationType.MODIFY_INFO,
            body, v2fields.LcmOperationStateType.PROCESSING)
        lcmocc.create(context)

        self.conductor_rpc.modify_vnfinfo(context, lcmocc.id)

        location = lcmocc_utils.lcmocc_href(lcmocc.id, self.endpoint)

        return sol_wsgi.SolResponse(202, None, location=location)

    @validator.schema(schema.InstantiateVnfRequest_V200, '2.0.0')
    @coordinate.lock_vnf_instance('{id}')
    def instantiate(self, request, id, body):
        context = request.context
        inst = inst_utils.get_inst(context, id)

        if inst.instantiationState != 'NOT_INSTANTIATED':
            raise sol_ex.VnfInstanceIsInstantiated(inst_id=id)

        lcmocc_utils.check_lcmocc_in_progress(context, id)

        lcmocc = self._new_lcmocc(id, v2fields.LcmOperationType.INSTANTIATE,
                                  body)

        req_param = lcmocc.operationParams
        # if there is partial vimConnectionInfo check and fulfill here.
        if req_param.obj_attr_is_set('vimConnectionInfo'):
            # if accessInfo or interfaceInfo is not specified, get from
            # VIM DB. vimId must be in VIM DB.
            req_vim_infos = req_param.vimConnectionInfo
            for key, req_vim_info in req_vim_infos.items():
                if not (req_vim_info.obj_attr_is_set('accessInfo') and
                        req_vim_info.obj_attr_is_set('interfaceInfo')):
                    vim_info = vim_utils.get_vim(context, req_vim_info.vimId)
                    req_vim_infos[key] = vim_info

        lcmocc.create(context)

        self.conductor_rpc.start_lcm_op(context, lcmocc.id)

        location = lcmocc_utils.lcmocc_href(lcmocc.id, self.endpoint)

        return sol_wsgi.SolResponse(202, None, location=location)

    @validator.schema(schema.TerminateVnfRequest_V200, '2.0.0')
    @coordinate.lock_vnf_instance('{id}')
    def terminate(self, request, id, body):
        context = request.context
        inst = inst_utils.get_inst(context, id)

        if inst.instantiationState != 'INSTANTIATED':
            raise sol_ex.VnfInstanceIsNotInstantiated(inst_id=id)

        lcmocc_utils.check_lcmocc_in_progress(context, id)

        lcmocc = self._new_lcmocc(id, v2fields.LcmOperationType.TERMINATE,
                                  body)
        lcmocc.create(context)

        self.conductor_rpc.start_lcm_op(context, lcmocc.id)

        location = lcmocc_utils.lcmocc_href(lcmocc.id, self.endpoint)

        return sol_wsgi.SolResponse(202, None, location=location)

    def _get_current_scale_level(self, inst, aspect_id):
        if (inst.obj_attr_is_set('instantiatedVnfInfo') and
                inst.instantiatedVnfInfo.obj_attr_is_set('scaleStatus')):
            for scale_info in inst.instantiatedVnfInfo.scaleStatus:
                if scale_info.aspectId == aspect_id:
                    return scale_info.scaleLevel

    def _get_max_scale_level(self, inst, aspect_id):
        if (inst.obj_attr_is_set('instantiatedVnfInfo') and
                inst.instantiatedVnfInfo.obj_attr_is_set('maxScaleLevels')):
            for scale_info in inst.instantiatedVnfInfo.maxScaleLevels:
                if scale_info.aspectId == aspect_id:
                    return scale_info.scaleLevel

    @validator.schema(schema.ScaleVnfRequest_V200, '2.0.0')
    @coordinate.lock_vnf_instance('{id}')
    def scale(self, request, id, body):
        context = request.context
        inst = inst_utils.get_inst(context, id)

        if inst.instantiationState != 'INSTANTIATED':
            raise sol_ex.VnfInstanceIsNotInstantiated(inst_id=id)

        lcmocc_utils.check_lcmocc_in_progress(context, id)

        # check parameters
        aspect_id = body['aspectId']
        if 'numberOfSteps' not in body:
            # set default value (1) defined by SOL specification for
            # the convenience of the following methods.
            body['numberOfSteps'] = 1

        scale_level = self._get_current_scale_level(inst, aspect_id)
        max_scale_level = self._get_max_scale_level(inst, aspect_id)
        if scale_level is None or max_scale_level is None:
            raise sol_ex.InvalidScaleAspectId(aspect_id=aspect_id)

        num_steps = body['numberOfSteps']
        if body['type'] == 'SCALE_IN':
            num_steps *= -1
        scale_level += num_steps
        if scale_level < 0 or scale_level > max_scale_level:
            raise sol_ex.InvalidScaleNumberOfSteps(
                num_steps=body['numberOfSteps'])

        lcmocc = self._new_lcmocc(id, v2fields.LcmOperationType.SCALE,
                                  body)
        lcmocc.create(context)

        self.conductor_rpc.start_lcm_op(context, lcmocc.id)

        location = lcmocc_utils.lcmocc_href(lcmocc.id, self.endpoint)

        return sol_wsgi.SolResponse(202, None, location=location)

    @validator.schema(schema.HealVnfRequest_V200, '2.0.0')
    @coordinate.lock_vnf_instance('{id}')
    def heal(self, request, id, body):
        context = request.context
        inst = inst_utils.get_inst(context, id)

        if inst.instantiationState != 'INSTANTIATED':
            raise sol_ex.VnfInstanceIsNotInstantiated(inst_id=id)

        lcmocc_utils.check_lcmocc_in_progress(context, id)

        # check parameter for later use
        is_all = body.get('additionalParams', {}).get('all', False)
        if not isinstance(is_all, bool):
            raise sol_ex.SolValidationError(
                detail="additionalParams['all'] must be bool.")

        if 'vnfcInstanceId' in body:
            inst_info = inst.instantiatedVnfInfo
            vnfc_id = []
            if inst_info.obj_attr_is_set('vnfcInfo'):
                vnfc_id = [vnfc.id for vnfc in inst_info.vnfcInfo]
            for req_vnfc_id in body['vnfcInstanceId']:
                if req_vnfc_id not in vnfc_id:
                    raise sol_ex.SolValidationError(
                        detail="vnfcInstanceId(%s) does not exist."
                        % req_vnfc_id)

        lcmocc = self._new_lcmocc(id, v2fields.LcmOperationType.HEAL, body)
        lcmocc.create(context)

        self.conductor_rpc.start_lcm_op(context, lcmocc.id)

        location = lcmocc_utils.lcmocc_href(lcmocc.id, self.endpoint)

        return sol_wsgi.SolResponse(202, None, location=location)

    @validator.schema(schema.ChangeExtVnfConnectivityRequest_V200, '2.0.0')
    @coordinate.lock_vnf_instance('{id}')
    def change_ext_conn(self, request, id, body):
        context = request.context
        inst = inst_utils.get_inst(context, id)

        if inst.instantiationState != 'INSTANTIATED':
            raise sol_ex.VnfInstanceIsNotInstantiated(inst_id=id)

        lcmocc_utils.check_lcmocc_in_progress(context, id)

        lcmocc = self._new_lcmocc(
            id, v2fields.LcmOperationType.CHANGE_EXT_CONN, body)
        lcmocc.create(context)

        self.conductor_rpc.start_lcm_op(context, lcmocc.id)

        location = lcmocc_utils.lcmocc_href(lcmocc.id, self.endpoint)

        return sol_wsgi.SolResponse(202, None, location=location)

    @validator.schema(schema.LccnSubscriptionRequest_V200, '2.0.0')
    def subscription_create(self, request, body):
        context = request.context
        subsc = objects.LccnSubscriptionV2(
            id=uuidutils.generate_uuid(),
            callbackUri=body['callbackUri'],
            verbosity=body.get('verbosity', 'FULL')  # default is 'FULL'
        )

        if body.get('filter'):
            subsc.filter = (
                objects.LifecycleChangeNotificationsFilterV2.from_dict(
                    body['filter'])
            )

        auth_req = body.get('authentication')
        if auth_req:
            auth = objects.SubscriptionAuthentication(
                authType=auth_req['authType']
            )
            if 'BASIC' in auth.authType:
                basic_req = auth_req.get('paramsBasic')
                if basic_req is None:
                    msg = "ParmasBasic must be specified."
                    raise sol_ex.InvalidSubscription(sol_detail=msg)
                auth.paramsBasic = (
                    objects.SubscriptionAuthentication_ParamsBasic(
                        userName=basic_req.get('userName'),
                        password=basic_req.get('password')
                    )
                )

            if 'OAUTH2_CLIENT_CREDENTIALS' in auth.authType:
                oauth2_req = auth_req.get('paramsOauth2ClientCredentials')
                if oauth2_req is None:
                    msg = "paramsOauth2ClientCredentials must be specified."
                    raise sol_ex.InvalidSubscription(sol_detail=msg)
                auth.paramsOauth2ClientCredentials = (
                    objects.SubscriptionAuthentication_ParamsOauth2(
                        clientId=oauth2_req.get('clientId'),
                        clientPassword=oauth2_req.get('clientPassword'),
                        tokenEndpoint=oauth2_req.get('tokenEndpoint')
                    )
                )

            if 'TLS_CERT' in auth.authType:
                msg = "'TLS_CERT' is not supported at the moment."
                raise sol_ex.InvalidSubscription(sol_detail=msg)

            subsc.authentication = auth

        if CONF.v2_nfvo.test_callback_uri:
            subsc_utils.test_notification(subsc)

        subsc.create(context)

        resp_body = self._subsc_view.detail(subsc)
        self_href = subsc_utils.subsc_href(subsc.id, self.endpoint)

        return sol_wsgi.SolResponse(201, resp_body, location=self_href)

    def subscription_list(self, request):
        filter_param = request.GET.get('filter')
        if filter_param is not None:
            filters = self._subsc_view.parse_filter(filter_param)
        else:
            filters = None

        subscs = subsc_utils.get_subsc_all(request.context)

        resp_body = self._subsc_view.detail_list(subscs, filters)

        return sol_wsgi.SolResponse(200, resp_body)

    def subscription_show(self, request, id):
        subsc = subsc_utils.get_subsc(request.context, id)

        resp_body = self._subsc_view.detail(subsc)

        return sol_wsgi.SolResponse(200, resp_body)

    def subscription_delete(self, request, id):
        context = request.context
        subsc = subsc_utils.get_subsc(request.context, id)

        subsc.delete(context)

        return sol_wsgi.SolResponse(204, None)

    def lcm_op_occ_list(self, request):
        filter_param = request.GET.get('filter')
        if filter_param is not None:
            filters = self._lcmocc_view.parse_filter(filter_param)
        else:
            filters = None

        selector = self._lcmocc_view.parse_selector(request.GET)

        lcmoccs = lcmocc_utils.get_lcmocc_all(request.context)

        resp_body = self._lcmocc_view.detail_list(lcmoccs, filters, selector)

        return sol_wsgi.SolResponse(200, resp_body)

    def lcm_op_occ_show(self, request, id):
        lcmocc = lcmocc_utils.get_lcmocc(request.context, id)

        resp_body = self._lcmocc_view.detail(lcmocc)

        return sol_wsgi.SolResponse(200, resp_body)

    def lcm_op_occ_retry(self, request, id):
        context = request.context
        lcmocc = lcmocc_utils.get_lcmocc(context, id)

        return self._lcm_op_occ_retry(context, lcmocc)

    @coordinate.lock_vnf_instance('{lcmocc.vnfInstanceId}')
    def _lcm_op_occ_retry(self, context, lcmocc):
        if lcmocc.operationState != v2fields.LcmOperationStateType.FAILED_TEMP:
            raise sol_ex.LcmOpOccNotFailedTemp(lcmocc_id=lcmocc.id)

        self.conductor_rpc.retry_lcm_op(context, lcmocc.id)

        return sol_wsgi.SolResponse(202, None)

    def lcm_op_occ_rollback(self, request, id):
        context = request.context
        lcmocc = lcmocc_utils.get_lcmocc(context, id)

        return self._lcm_op_occ_rollback(context, lcmocc)

    @coordinate.lock_vnf_instance('{lcmocc.vnfInstanceId}')
    def _lcm_op_occ_rollback(self, context, lcmocc):
        if lcmocc.operationState != v2fields.LcmOperationStateType.FAILED_TEMP:
            raise sol_ex.LcmOpOccNotFailedTemp(lcmocc_id=lcmocc.id)

        self.conductor_rpc.rollback_lcm_op(context, lcmocc.id)

        return sol_wsgi.SolResponse(202, None)

    def lcm_op_occ_fail(self, request, id):
        context = request.context
        lcmocc = lcmocc_utils.get_lcmocc(context, id)

        return self._lcm_op_occ_fail(context, lcmocc)

    @coordinate.lock_vnf_instance('{lcmocc.vnfInstanceId}')
    def _lcm_op_occ_fail(self, context, lcmocc):
        if lcmocc.operationState != v2fields.LcmOperationStateType.FAILED_TEMP:
            raise sol_ex.LcmOpOccNotFailedTemp(lcmocc_id=lcmocc.id)

        inst = inst_utils.get_inst(context, lcmocc.vnfInstanceId)
        grant_req, grant = lcmocc_utils.get_grant_req_and_grant(context,
                                                                lcmocc)

        lcmocc.operationState = v2fields.LcmOperationStateType.FAILED
        with context.session.begin(subtransactions=True):
            lcmocc.update(context)
            if grant_req is not None:
                grant_req.delete(context)
                grant.delete(context)

        # send notification FAILED
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

        resp_body = self._lcmocc_view.detail(lcmocc)

        return sol_wsgi.SolResponse(200, resp_body)

    def lcm_op_occ_delete(self, request, id):
        # not allowed to delete on the specification
        if not CONF.v2_vnfm.test_enable_lcm_op_occ_delete:
            raise sol_ex.MethodNotAllowed(method='DELETE')

        # NOTE: This is for test use since it is inconvenient not to be
        # able to delete.
        context = request.context
        lcmocc = lcmocc_utils.get_lcmocc(context, id)

        lcmocc.delete(context)

        return sol_wsgi.SolResponse(204, None)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import testtools
from unittest import mock

from tacker.vnfm.monitor_drivers.ping import ping


class TestVNFMonitorPing(testtools.TestCase):

    def setUp(self):
        super(TestVNFMonitorPing, self).setUp()
        self.monitor_ping = ping.VNFMonitorPing()

    @mock.patch('tacker.agent.linux.utils.execute')
    def test_monitor_call_for_success(self, mock_utils_execute):
        test_vnf = {}
        test_kwargs = {
            'mgmt_ip': 'a.b.c.d'
        }
        mock_ping_cmd = ['ping',
                         '-c', 5,
                         '-W', 5.0,
                         '-i', 1.0,
                         'a.b.c.d']
        self.monitor_ping.monitor_call(test_vnf,
                                       test_kwargs)
        mock_utils_execute.assert_called_once_with(mock_ping_cmd,
                                                   check_exit_code=True)

    @mock.patch('tacker.agent.linux.utils.execute')
    def test_monitor_call_for_failure(self, mock_utils_execute):
        mock_utils_execute.side_effect = RuntimeError()
        test_vnf = {}
        test_kwargs = {
            'mgmt_ip': 'a.b.c.d'
        }
        monitor_return = self.monitor_ping.monitor_call(test_vnf,
                                                        test_kwargs)
        self.assertEqual('failure', monitor_return)

    def test_monitor_url(self):
        test_vnf = {
            'monitor_url': 'a.b.c.d'
        }
        test_monitor_url = self.monitor_ping.monitor_url(mock.ANY,
                                                         mock.ANY,
                                                         test_vnf)
        self.assertEqual('a.b.c.d', test_monitor_url)

    @mock.patch('tacker.agent.linux.utils.execute')
    def test_monitor_call_with_params(self, mock_utils_execute):
        check_ping_counts = 2
        check_ping_timeout = 5.0
        check_ping_interval = 0.5
        test_vnf = {}
        test_kwargs = {
            'mgmt_ip': 'a:b:c:d:e:f:1:2',
            'count': check_ping_counts,
            'timeout': check_ping_timeout,
            'interval': check_ping_interval
        }
        mock_ping_cmd = ['ping6',
                         '-c', check_ping_counts,
                         '-W', check_ping_timeout,
                         '-i', check_ping_interval,
                         'a:b:c:d:e:f:1:2']
        self.monitor_ping.monitor_call(test_vnf,
                                       test_kwargs)
        mock_utils_execute.assert_called_once_with(mock_ping_cmd,
                                                   check_exit_code=True)

    @mock.patch('tacker.agent.linux.utils.execute')
    def test_monitor_call_for_counts(self, mock_utils_execute):
        check_retury_counts = 5
        mock_utils_execute.side_effect = RuntimeError()
        test_vnf = {}
        test_kwargs = {
            'mgmt_ip': 'a:b:c:d:e:f:1:2',
            'retry': check_retury_counts
        }
        self.monitor_ping.monitor_call(test_vnf,
                                       test_kwargs)
        self.assertEqual(check_retury_counts,
                         mock_utils_execute.call_count)



# Copyright 2011 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Tacker
"""

import os

from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging
import oslo_messaging
from paste import deploy

from tacker._i18n import _
from tacker.common import utils
from tacker import version


LOG = logging.getLogger(__name__)

core_opts = [
    cfg.HostAddressOpt('bind_host', default='0.0.0.0',
                       help=_("The host IP to bind to")),
    cfg.IntOpt('bind_port', default=9890,
               help=_("The port to bind to")),
    cfg.StrOpt('api_paste_config', default="api-paste.ini",
               help=_("The API paste config file to use")),
    cfg.StrOpt('api_extensions_path', default="",
               help=_("The path for API extensions")),
    cfg.ListOpt('service_plugins', default=['nfvo', 'vnfm'],
                help=_("The service plugins Tacker will use")),
    cfg.StrOpt('auth_strategy', default='keystone',
               help=_("The type of authentication to use")),
    cfg.BoolOpt('allow_bulk', default=True,
                help=_("Allow the usage of the bulk API")),
    cfg.BoolOpt('allow_pagination', default=False,
                help=_("Allow the usage of the pagination")),
    cfg.BoolOpt('allow_sorting', default=False,
                help=_("Allow the usage of the sorting")),
    cfg.StrOpt('pagination_max_limit', default="-1",
               help=_("The maximum number of items returned "
                      "in a single response, value was 'infinite' "
                      "or negative integer means no limit")),
    cfg.HostAddressOpt('host', default=utils.get_hostname(),
                       help=_("The hostname Tacker is running on")),
]

core_cli_opts = [
    cfg.StrOpt('state_path',
               default='/var/lib/tacker',
               help=_("Where to store Tacker state files. "
                      "This directory must be writable by "
                      "the agent.")),
]

logging.register_options(cfg.CONF)
# Register the configuration options
cfg.CONF.register_opts(core_opts)
cfg.CONF.register_cli_opts(core_cli_opts)


def config_opts():
    return [(None, core_opts), (None, core_cli_opts)]


# Ensure that the control exchange is set correctly
oslo_messaging.set_transport_defaults(control_exchange='tacker')


def set_db_defaults():
    # Update the default QueuePool parameters. These can be tweaked by the
    # conf variables - max_pool_size, max_overflow and pool_timeout
    db_options.set_defaults(
        cfg.CONF,
        connection='sqlite://',
        max_pool_size=10,
        max_overflow=20, pool_timeout=10)


set_db_defaults()


def init(args, **kwargs):
    cfg.CONF(args=args, project='tacker',
             version='%%prog %s' % version.version_info.release_string(),
             **kwargs)

    # FIXME(ihrachys): if import is put in global, circular import
    # failure occurs
    from tacker.common import rpc as n_rpc
    n_rpc.init(cfg.CONF)


def setup_logging(conf):
    """Sets up the logging options for a log with supplied name.

    :param conf: a cfg.ConfOpts object
    """
    product_name = "tacker"
    logging.setup(conf, product_name)
    LOG.info("Logging enabled!")


def load_paste_app(app_name):
    """Builds and returns a WSGI app from a paste config file.

    :param app_name: Name of the application to load
    :raises ConfigFilesNotFoundError: when config file cannot be located
    :raises RuntimeError: when application cannot be loaded from config file
    """

    config_path = cfg.CONF.find_file(cfg.CONF.api_paste_config)
    if not config_path:
        raise cfg.ConfigFilesNotFoundError(
            config_files=[cfg.CONF.api_paste_config])
    config_path = os.path.abspath(config_path)
    LOG.info("Config paste file: %s", config_path)

    try:
        app = deploy.loadapp("config:%s" % config_path, name=app_name)
    except (LookupError, ImportError):
        msg = (_("Unable to load %(app_name)s from "
                 "configuration file %(config_path)s.") %
               {'app_name': app_name,
                'config_path': config_path})
        LOG.exception(msg)
        raise RuntimeError(msg)
    return app



# Copyright (C) 2021 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import ipaddress
import os
import time

import eventlet
from oslo_log import log as logging
import paramiko

from tacker.common import cmd_executer
from tacker.common import exceptions
from tacker import objects
from tacker.vnflcm import utils as vnflcm_utils
from tacker.vnfm.infra_drivers.openstack import heat_client as hc
from tacker.vnfm.mgmt_drivers import vnflcm_abstract_driver

LOG = logging.getLogger(__name__)

# CLI timeout period
PR_CONNECT_TIMEOUT = 30
PR_CMD_TIMEOUT_DEFAULT = 600
PR_CMD_TIMEOUT_INSTALL = 2700

# retry interval(sec)
PR_CMD_RETRY_INTERVAL = 30

# number of check command retries for wait for Docker running
PR_NUM_OF_RETRY_WAIT_DOCKER = 5

# number of check command retries for wait for Private registry running
PR_NUM_OF_RETRY_WAIT_PR = 5

# Command type
CMD_TYPE_COMMON = "common"

# Default host port
DEFAULT_HOST_PORT = '5000'


class PrivateRegistryMgmtDriver(
        vnflcm_abstract_driver.VnflcmMgmtAbstractDriver):

    def get_type(self):
        return "mgmt-drivers-private-registry"

    def get_name(self):
        return "mgmt-drivers-private-registry"

    def get_description(self):
        return "Tacker Private registry VNF Mgmt Driver"

    def _get_cp_ip_address(self, vnf_instance, vim_connection_info, cp_name):
        heatclient = hc.HeatClient(vim_connection_info.access_info)
        stack_id = vnf_instance.instantiated_vnf_info.instance_id

        # get IP address from heat
        resource_info = heatclient.resources.get(
            stack_id=stack_id, resource_name=cp_name)
        cp_ip_address = resource_info.attributes.get('floating_ip_address')
        if cp_ip_address is None and resource_info.attributes.get('fixed_ips'):
            cp_ip_address = resource_info.attributes.get(
                'fixed_ips')[0].get("ip_address")

        # check result
        try:
            ipaddress.ip_address(cp_ip_address)
        except ValueError:
            err_msg = "The IP address of Private registry VM is invalid."
            LOG.error(err_msg)
            raise exceptions.MgmtDriverOtherError(error_message=err_msg)
        if cp_ip_address is None:
            err_msg = "Failed to get IP address for Private registry VM"
            LOG.error(err_msg)
            raise exceptions.MgmtDriverOtherError(error_message=err_msg)

        LOG.debug("Getting IP address succeeded. "
            "(CP name: {}, IP address: {})".format(cp_name, cp_ip_address))
        return cp_ip_address

    def _execute_command(self, commander, ssh_command,
                         timeout=PR_CMD_TIMEOUT_DEFAULT,
                         type=CMD_TYPE_COMMON, retry=0):
        eventlet.monkey_patch()
        while retry >= 0:
            try:
                with eventlet.Timeout(timeout, True):
                    LOG.debug("execute command: {}".format(ssh_command))
                    result = commander.execute_command(
                        ssh_command, input_data=None)
                    break
            except eventlet.timeout.Timeout:
                err_msg = ("It is time out, When execute command: "
                    "{}.".format(ssh_command))
                retry -= 1
                if retry < 0:
                    LOG.error(err_msg)
                    commander.close_session()
                    raise exceptions.MgmtDriverOtherError(
                        error_message=err_msg)
                err_msg += " Retry after {} seconds.".format(
                    PR_CMD_RETRY_INTERVAL)
                LOG.debug(err_msg)
                time.sleep(PR_CMD_RETRY_INTERVAL)
        if type == CMD_TYPE_COMMON:
            stderr = result.get_stderr()
            if stderr:
                err_msg = ("Failed to execute command: {}, "
                    "stderr: {}".format(ssh_command, stderr))
                LOG.error(err_msg)
                commander.close_session()
                raise exceptions.MgmtDriverOtherError(error_message=err_msg)
        return result.get_stdout()

    def _wait_docker_running(self, commander, err_msg,
                             retry=PR_NUM_OF_RETRY_WAIT_DOCKER):
        while retry >= 0:
            ssh_command = ("sudo systemctl status docker "
                "| grep Active | grep -c running")
            result = self._execute_command(commander, ssh_command)
            count_result = result[0].replace("\n", "")
            if count_result == "0":
                retry -= 1
                if retry < 0:
                    LOG.error(err_msg)
                    commander.close_session()
                    raise exceptions.MgmtDriverOtherError(
                        error_message=err_msg)
                LOG.debug("Docker service is not running. "
                    "Check again after {} seconds.".format(
                        PR_CMD_RETRY_INTERVAL))
                time.sleep(PR_CMD_RETRY_INTERVAL)
            else:
                LOG.debug("Docker service is running.")
                break

    def _wait_private_registry_running(self, commander,
                                       retry=PR_NUM_OF_RETRY_WAIT_PR):
        while retry >= 0:
            ssh_command = ("sudo docker inspect "
                "--format=\'{{.State.Status}}\' "
                "private_registry")
            result = self._execute_command(commander, ssh_command)
            status = result[0].replace("\n", "")
            if status == "running":
                LOG.debug("Private registry container is running.")
                break
            retry -= 1
            if retry < 0:
                err_msg = "Failed to run Private registry container"
                LOG.error(err_msg)
                commander.close_session()
                raise exceptions.MgmtDriverOtherError(
                    error_message=err_msg)
            LOG.debug("Private registry container is not running. "
                "Check again after {} seconds.".format(
                    PR_CMD_RETRY_INTERVAL))
            time.sleep(PR_CMD_RETRY_INTERVAL)

    def _check_pr_installation_params(self, pr_installation_params):
        if not pr_installation_params:
            LOG.error("The private_registry_installation_param "
                "in the additionalParams does not exist.")
            raise exceptions.MgmtDriverNotFound(
                param="private_registry_installation_param")
        ssh_cp_name = pr_installation_params.get("ssh_cp_name")
        ssh_username = pr_installation_params.get("ssh_username")
        ssh_password = pr_installation_params.get("ssh_password")
        if not ssh_cp_name:
            LOG.error("The ssh_cp_name "
                "in the additionalParams does not exist.")
            raise exceptions.MgmtDriverNotFound(param="ssh_cp_name")
        if not ssh_username:
            LOG.error("The ssh_username "
                "in the additionalParams does not exist.")
            raise exceptions.MgmtDriverNotFound(param="ssh_username")
        if not ssh_password:
            LOG.error("The ssh_password "
                "in the additionalParams does not exist.")
            raise exceptions.MgmtDriverNotFound(param="ssh_password")

    def _install_private_registry(self, context, vnf_instance,
                                  vim_connection_info,
                                  pr_installation_params):
        LOG.debug("Start private registry installation. "
            "installation param: {}".format(pr_installation_params))

        # check parameters
        self._check_pr_installation_params(pr_installation_params)

        ssh_cp_name = pr_installation_params.get("ssh_cp_name")
        ssh_username = pr_installation_params.get("ssh_username")
        ssh_password = pr_installation_params.get("ssh_password")
        image_path = pr_installation_params.get("image_path")
        port_no = pr_installation_params.get("port_no")
        proxy = pr_installation_params.get("proxy")

        # get IP address from cp name
        ssh_ip_address = self._get_cp_ip_address(
            vnf_instance, vim_connection_info, ssh_cp_name)

        # initialize RemoteCommandExecutor
        retry = 4
        while retry > 0:
            try:
                commander = cmd_executer.RemoteCommandExecutor(
                    user=ssh_username, password=ssh_password,
                    host=ssh_ip_address, timeout=PR_CONNECT_TIMEOUT)
                break
            except (exceptions.NotAuthorized, paramiko.SSHException,
                    paramiko.ssh_exception.NoValidConnectionsError) as e:
                LOG.debug(e)
                retry -= 1
                if retry < 0:
                    err_msg = "Failed to use SSH to connect to the registry " \
                              "server: {}".format(ssh_ip_address)
                    LOG.error(err_msg)
                    raise exceptions.MgmtDriverOtherError(
                        error_message=err_msg)
                time.sleep(PR_CMD_RETRY_INTERVAL)

        # check OS and architecture
        ssh_command = ("cat /etc/os-release "
            "| grep \"PRETTY_NAME=\" "
            "| grep -c \"Ubuntu 20.04\"; arch | grep -c x86_64")
        result = self._execute_command(commander, ssh_command)
        os_check_result = result[0].replace("\n", "")
        arch_check_result = result[1].replace("\n", "")
        if os_check_result == "0" or arch_check_result == "0":
            err_msg = ("Failed to install. "
                "Your OS does not support at present. "
                "It only supports Ubuntu 20.04 (x86_64)")
            LOG.error(err_msg)
            commander.close_session()
            raise exceptions.MgmtDriverOtherError(error_message=err_msg)

        # get proxy params
        http_proxy = ""
        https_proxy = ""
        no_proxy = ""
        if proxy:
            http_proxy = proxy.get("http_proxy")
            https_proxy = proxy.get("https_proxy")
            no_proxy = proxy.get("no_proxy")

        # execute apt-get install command
        ssh_command = ""
        if http_proxy or https_proxy:
            # set apt's proxy config
            ssh_command = "echo -e \""
            if http_proxy:
                ssh_command += ("Acquire::http::Proxy "
                    "\\\"{}\\\";\\n".format(http_proxy))
            if https_proxy:
                ssh_command += ("Acquire::https::Proxy "
                    "\\\"{}\\\";\\n".format(https_proxy))
            ssh_command += ("\" | sudo tee /etc/apt/apt.conf.d/proxy.conf "
                ">/dev/null && ")
        ssh_command += (
            "sudo apt-get update && "
            "export DEBIAN_FRONTEND=noninteractive;"
            "sudo -E apt-get install -y apt-transport-https "
            "ca-certificates curl gnupg-agent software-properties-common")
        self._execute_command(commander, ssh_command, PR_CMD_TIMEOUT_INSTALL)

        # execute add-apt-repository command
        ssh_command = ""
        if http_proxy:
            ssh_command += "export http_proxy=\"{}\";".format(http_proxy)
        if https_proxy:
            ssh_command += "export https_proxy=\"{}\";".format(https_proxy)
        if no_proxy:
            ssh_command += "export no_proxy=\"{}\";".format(no_proxy)
        ssh_command += (
            "export APT_KEY_DONT_WARN_ON_DANGEROUS_USAGE=DontWarn;"
            "curl -fsSL https://download.docker.com/linux/ubuntu/gpg "
            "| sudo -E apt-key add - && "
            "sudo add-apt-repository \"deb [arch=amd64] "
            "https://download.docker.com/linux/ubuntu "
            "$(lsb_release -cs) stable\"")
        self._execute_command(commander, ssh_command, PR_CMD_TIMEOUT_INSTALL)

        # install docker
        ssh_command = (
            "sudo apt-get update && "
            "export DEBIAN_FRONTEND=noninteractive;"
            "sudo -E apt-get install -y "
            "docker-ce=5:19.03.11~3-0~ubuntu-focal "
            "docker-ce-cli containerd.io")
        self._execute_command(commander, ssh_command, PR_CMD_TIMEOUT_INSTALL)

        # wait for the Docker service running
        err_msg = "Failed to install Docker(Docker service is not running)"
        self._wait_docker_running(commander, err_msg)

        # set Docker's proxy config
        if http_proxy or https_proxy or no_proxy:
            proxy_env_list = []
            if http_proxy:
                proxy_env = "\\\"HTTP_PROXY={}\\\"".format(http_proxy)
                proxy_env_list.append(proxy_env)
            if https_proxy:
                proxy_env = "\\\"HTTPS_PROXY={}\\\"".format(https_proxy)
                proxy_env_list.append(proxy_env)
            if no_proxy:
                proxy_env = "\\\"NO_PROXY={}\\\"".format(no_proxy)
                proxy_env_list.append(proxy_env)
            proxy_env = " ".join(proxy_env_list)
            ssh_command = (
                "sudo mkdir -p /etc/systemd/system/docker.service.d && "
                "echo -e \"[Service]\\nEnvironment={}\" | sudo tee "
                "/etc/systemd/system/docker.service.d/https-proxy.conf "
                ">/dev/null && "
                "sudo systemctl daemon-reload && "
                "sudo systemctl restart docker".format(proxy_env))
            self._execute_command(commander, ssh_command)

            # wait for the Docker service running
            err_msg = ("Failed to restart Docker"
                "(Docker service is not running)")
            self._wait_docker_running(commander, err_msg)

        # pull or load the Docker image named "registry"
        if not image_path:
            # pull the Docker image
            ssh_command = "sudo docker pull registry"
            self._execute_command(commander, ssh_command)
        else:
            vnf_package_path = vnflcm_utils._get_vnf_package_path(
                context, vnf_instance.vnfd_id)
            local_image_path = os.path.join(
                vnf_package_path, image_path)

            # check existence of local image file
            if not os.path.exists(local_image_path):
                LOG.error("The image_path in the additionalParams is invalid. "
                    "File does not exist.")
                commander.close_session()
                raise exceptions.MgmtDriverParamInvalid(param="image_path")

            # transfer the Docker image file to Private registry VM
            image_file_name = os.path.basename(image_path)
            remote_image_path = os.path.join("/tmp", image_file_name)
            transport = paramiko.Transport(ssh_ip_address, 22)
            transport.connect(username=ssh_username, password=ssh_password)
            sftp_client = paramiko.SFTPClient.from_transport(transport)
            sftp_client.put(local_image_path, remote_image_path)
            transport.close()

            # load the Docker image
            ssh_command = "sudo docker load -i {}".format(remote_image_path)
            self._execute_command(commander, ssh_command)

        # check Docker images list
        ssh_command = "sudo docker images | grep -c registry"
        result = self._execute_command(commander, ssh_command)
        count_result = result[0].replace("\n", "")
        if count_result == "0":
            err_msg = "Failed to pull or load the Docker image named registry"
            LOG.error(err_msg)
            commander.close_session()
            raise exceptions.MgmtDriverOtherError(error_message=err_msg)

        # run the Private registry container
        if port_no is None:
            port = DEFAULT_HOST_PORT
        else:
            port = str(port_no)
        ssh_command = (
            "sudo docker run -d -p {}:5000 "
            "-v /private_registry:/var/lib/registry "
            "--restart=always "
            "--name private_registry "
            "registry:latest".format(port))
        self._execute_command(commander, ssh_command)

        # wait for the Private registry container running
        self._wait_private_registry_running(commander)

        commander.close_session()
        LOG.debug("Private registry installation complete.")

    def instantiate_start(self, context, vnf_instance,
                          instantiate_vnf_request, grant,
                          grant_request, **kwargs):
        pass

    def instantiate_end(self, context, vnf_instance,
                        instantiate_vnf_request, grant,
                        grant_request, **kwargs):
        # get vim_connection_info
        vim_info = vnflcm_utils._get_vim(context,
            instantiate_vnf_request.vim_connection_info)
        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        # get parameters for private registry installation
        pr_installation_params = instantiate_vnf_request.additional_params.get(
            "private_registry_installation_param")

        self._install_private_registry(
            context, vnf_instance, vim_connection_info, pr_installation_params)

    def terminate_start(self, context, vnf_instance,
                        terminate_vnf_request, grant,
                        grant_request, **kwargs):
        pass

    def terminate_end(self, context, vnf_instance,
                      terminate_vnf_request, grant,
                      grant_request, **kwargs):
        pass

    def scale_start(self, context, vnf_instance,
                    scale_vnf_request, grant,
                    grant_request, **kwargs):
        pass

    def scale_end(self, context, vnf_instance,
                  scale_vnf_request, grant,
                  grant_request, **kwargs):
        pass

    def heal_start(self, context, vnf_instance,
                   heal_vnf_request, grant,
                   grant_request, **kwargs):
        pass

    def heal_end(self, context, vnf_instance,
                 heal_vnf_request, grant,
                 grant_request, **kwargs):
        # NOTE: Private registry VNF has only one VNFC.
        # Therefore, VNFC that is repaired by entire Heal and
        # VNFC that is repaired by specifying VNFC instance are the same VNFC.

        # get vim_connection_info
        vim_info = vnflcm_utils._get_vim(context,
            vnf_instance.vim_connection_info)
        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        # get parameters for private registry installation
        pr_installation_params = (
            vnf_instance.instantiated_vnf_info.additional_params.get(
                "private_registry_installation_param"))

        self._install_private_registry(
            context, vnf_instance, vim_connection_info, pr_installation_params)

    def change_external_connectivity_start(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass

    def change_external_connectivity_end(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass

    def modify_information_start(self, context, vnf_instance,
                                 modify_vnf_request, **kwargs):
        pass

    def modify_information_end(self, context, vnf_instance,
                               modify_vnf_request, **kwargs):
        pass



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker import context
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase


class TestTerminateVnfRequest(SqlTestCase):

    def setUp(self):
        super(TestTerminateVnfRequest, self).setUp()
        self.context = context.get_admin_context()

    def _get_terminate_vnf_request(self):
        terminate_vnf_request = {
            'termination_type': 'GRACEFUL',
            'graceful_termination_timeout': 10
        }
        return terminate_vnf_request

    def test_obj_from_primitive(self):
        terminate_vnf_request = self._get_terminate_vnf_request()
        result = objects.TerminateVnfRequest.obj_from_primitive(
            terminate_vnf_request, self.context)
        self.assertIsInstance(result, objects.TerminateVnfRequest)
        self.assertEqual('GRACEFUL', result.termination_type)
        self.assertEqual(terminate_vnf_request['graceful_termination_timeout'],
            result.graceful_termination_timeout)

    def test_obj_from_primitive_without_timeout(self):
        terminate_vnf_request = self._get_terminate_vnf_request()
        terminate_vnf_request.pop('graceful_termination_timeout')

        result = objects.TerminateVnfRequest.obj_from_primitive(
            terminate_vnf_request, self.context)
        self.assertIsInstance(result, objects.TerminateVnfRequest)
        self.assertEqual('GRACEFUL', result.termination_type)
        self.assertEqual(0, result.graceful_termination_timeout)
